from __future__ import annotations

import uuid
from datetime import datetime
from enum import Enum

from sqlalchemy.orm import validates

from meltano.core import utils
from meltano.core.sqlalchemy import GUID

from . import db


class SubscriptionEventType(str, Enum):
    PIPELINE_MANUAL_RUN = "pipeline_manual_run"


class Subscription(db.Model):
    __tablename__ = "subscriptions"

    id = db.Column(GUID, primary_key=True, default=uuid.uuid4)
    recipient = db.Column(db.String(), nullable=False)
    event_type = db.Column(db.String(), nullable=False)
    source_type = db.Column(db.String(), nullable=True)
    source_id = db.Column(db.String(), nullable=True)
    created_at = db.Column(db.DateTime(), default=datetime.utcnow)

    @validates("recipient")
    def validate_recipient(self, key, recipient):
        if not utils.is_email_valid(recipient):
            raise AssertionError(
                f"Validation failed: `{key}` must be a valid email address."
            )

        return recipient



"""Tracking plugin context for the Snowplow tracker."""

from __future__ import annotations

import uuid

from snowplow_tracker import SelfDescribingJson
from structlog.stdlib import get_logger

from meltano.core.block.blockset import BlockSet
from meltano.core.block.plugin_command import PluginCommandBlock
from meltano.core.elt_context import ELTContext
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.tracking.schemas import PluginsContextSchema
from meltano.core.utils import hash_sha256, safe_hasattr

logger = get_logger(__name__)


def _from_plugin(plugin: ProjectPlugin, cmd: str | None) -> dict:
    if not plugin or not safe_hasattr(plugin, "info"):
        # don't try to snag any info for this plugin, we're somehow badly malformed (unittest?), or where passed None.
        # this event will be routed to the "bad" bucket on the snowplow side. That makes it detectable on our end,
        # unlike if we had just filtered it out completely.
        logger.debug(
            "Plugin tracker context some how encountered plugin without info attr."
        )
        return {}

    return {
        "category": str(plugin.type),
        "name_hash": hash_sha256(plugin.name) if plugin.name else None,
        "namespace_hash": hash_sha256(plugin.namespace) if plugin.namespace else None,
        "executable_hash": hash_sha256(plugin.executable)
        if plugin.executable
        else None,
        "variant_name_hash": hash_sha256(plugin.variant) if plugin.variant else None,
        "pip_url_hash": hash_sha256(plugin.pip_url) if plugin.pip_url else None,
        "parent_name_hash": hash_sha256(plugin.parent.name) if plugin.parent else None,
        "command": cmd,
    }


# Proactively named to avoid name collisions more widely used "PluginContext"
class PluginsTrackingContext(SelfDescribingJson):
    """Tracking context for the Meltano plugins."""

    def __init__(self, plugins: list(tuple[ProjectPlugin, str | None])):
        """Initialize a meltano tracking plugin context.

        Args:
            plugins: The Meltano plugins and the requested command.
        """
        super().__init__(
            PluginsContextSchema.url,
            {
                "context_uuid": str(uuid.uuid4()),
                "plugins": [_from_plugin(plugin, cmd) for plugin, cmd in plugins],
            },
        )

    @classmethod
    def from_elt_context(cls, elt_context: ELTContext) -> PluginsTrackingContext:
        """Create a PluginsTrackingContext from an ELTContext.

        Args:
            elt_context: The ELTContext to use.

        Returns:
            A PluginsTrackingContext.
        """
        plugins = []
        if not elt_context.only_transform:
            plugins.extend(
                (
                    (elt_context.extractor.plugin, None),
                    (elt_context.loader.plugin, None),
                )
            )
        if elt_context.transformer:
            plugins.append((elt_context.transformer.plugin, None))
        return cls(plugins)

    @classmethod
    def from_block(cls, blk: BlockSet | PluginCommandBlock) -> PluginsTrackingContext:
        """Create a PluginsTrackingContext from a BlockSet or PluginCommandBlock.

        Args:
            blk: The block to create the context for.

        Raises:
            TypeError: `blk` is not a `BlockSet` or `PluginCommandBlock`.

        Returns:
            The PluginsTrackingContext for the given block.
        """
        if isinstance(blk, BlockSet):
            plugins: list[tuple[ProjectPlugin, str]] = [
                (plugin_block.context.plugin, plugin_block.plugin_args)
                for plugin_block in blk.blocks
            ]

            return cls(plugins)
        if isinstance(blk, PluginCommandBlock):
            return cls([(blk.context.plugin, blk.command)])
        raise TypeError(
            "Parameter 'blk' must be an instance of 'BlockSet' or 'PluginCommandBlock', "
            + f"not {type(blk)!r}"
        )

    @classmethod
    def from_blocks(
        cls, parsed_blocks: list[BlockSet | PluginCommandBlock]
    ) -> PluginsTrackingContext:
        """Create a PluginsTrackingContext from a list of BlockSets or PluginCommandBlocks.

        Args:
            parsed_blocks: The blocks to create the context from.

        Returns:
            The PluginsTrackingContext for the given blocks.
        """
        plugins: list[tuple[ProjectPlugin, str]] = []
        for blk in parsed_blocks:
            if isinstance(blk, BlockSet):
                plugins.extend(
                    (plugin_block.context.plugin, plugin_block.plugin_args)
                    for plugin_block in blk.blocks
                )
            elif isinstance(blk, PluginCommandBlock):
                plugins.append((blk.context.plugin, blk.command))
        return PluginsTrackingContext(plugins)



from __future__ import annotations

from pathlib import Path

from .elt_context import ELTContext


class DialectNotSupportedError(Exception):
    """Exception for when a dialect is not supported."""


class ConnectionService:
    def __init__(self, elt_context: ELTContext):
        self.context = elt_context

    @property
    def dialect(self):
        return self.context.loader.get_config("_dialect")

    def analyze_params(self):
        if self.dialect not in {"postgres", "snowflake"}:
            return {}

        if self.context.transformer:
            schema = self.context.transformer.get_config("target_schema")
        else:
            schema = self.context.loader.get_config("_target_schema")

        return {"schema": schema}

    def analyze_uri(self):
        params = self.analyze_params()
        return self.dialect_engine_uri(params)

    def dialect_engine_uri(self, params=None):
        params = {} if params is None else params

        # this part knows probably too much about the setting definition
        # for a plugin, but we have to do this if we want to infer connections
        # from the loader

        def db_suffix(s):
            return str(Path(s).with_suffix(".db"))

        dialect_templates = {
            "postgres": lambda params: "postgresql://{user}:{password}@{host}:{port}/{dbname}".format(
                **params
            ),
            "snowflake": lambda params: "snowflake://{username}:{password}@{account}/{database}/{schema}?warehouse={warehouse}&role={role}".format(
                **params
            ),
            "sqlite": lambda params: "sqlite:///{database}".format(
                database=db_suffix(params.pop("database")), **params
            ),
        }

        try:
            return dialect_templates[self.dialect](
                {**self.context.loader.config_dict(), **params}
            )
        except KeyError:
            raise DialectNotSupportedError(self.dialect)



from __future__ import annotations

import logging
from enum import Enum

import humps
from flask import current_app, json, request

from meltano.core.utils import compose


class KeyFrozenDict(dict):
    """Dict subclass that marks the dict as "frozen"."""


def freeze_keys(d: dict):
    """Mark the dictionary as frozen - no automatic conversion will be operated on it."""
    if isinstance(d, KeyFrozenDict):
        return d

    return KeyFrozenDict(d)


class JSONScheme(str, Enum):
    CAMEL_CASE = "camel"
    SNAKE_CASE = "snake"


def key_convert(obj, converter):
    if isinstance(obj, dict) and not isinstance(obj, KeyFrozenDict):
        converted = {}
        for k, v in obj.items():
            # humps fails to convert undescored values
            # see https://github.com/nficano/humps/issues/2
            if k.startswith("_"):
                new_k = k
            else:
                new_k = converter(k)

            if new_k in converted:
                raise ValueError(f"Naming scheme conversion conflict on `{new_k}`")

            converted[new_k] = key_convert(v, converter)

        return converted
    elif isinstance(obj, list):
        return [key_convert(x, converter) for x in obj]
    return obj


class JSONSchemeDecoder(json.JSONDecoder):
    """This JSONDecoder normalizes the returned `dict` to use `snake` case naming scheme."""

    def __init__(self, *args, **kwargs):
        hooks = compose(kwargs.pop("object_hook", None), self.hook)

        super().__init__(*args, **kwargs, object_hook=hooks)

    def hook(self, obj):
        # transform to snakecase if possible
        try:
            return key_convert(obj, humps.decamelize)
        except ValueError as err:
            logging.err(str(err))
            return super().hook(obj)


class JSONSchemeEncoder(json.JSONEncoder):
    """This JSONEncoder supports the `X-JSON-SCHEME` header to send the payload using requested naming scheme.

    Available schemes are `camel` and `snake`.
    Defaults to `snake`.
    """

    case_strategies = {
        JSONScheme.CAMEL_CASE: humps.camelize,
        JSONScheme.SNAKE_CASE: humps.decamelize,
    }

    def encode(self, obj):
        header = current_app.config["JSON_SCHEME_HEADER"]

        try:
            scheme = request.headers.get(header)
            strategy = self.__class__.case_strategies[scheme]
            logging.debug(f"Using JSON Scheme: {scheme}")
            obj = key_convert(obj, strategy)
        except ValueError as err:
            logging.error(str(err))
            # if we can't convert the keys, let's fallback to
            # the default implementation
            return super().encode(obj)
        except KeyError:
            pass

        return super().encode(obj)


def setup_json(app):
    app.json_encoder = JSONSchemeEncoder
    app.json_decoder = JSONSchemeDecoder

    # flask-restful uses its own encoder
    app.config["RESTFUL_JSON"]["cls"] = app.json_encoder



"""Various utils and formatters for log rendering control."""

from __future__ import annotations

from typing import Callable, Sequence, TextIO

import click
import structlog
from rich.console import Console
from rich.traceback import Traceback, install
from structlog.types import Processor

from meltano.core.utils import get_no_color_flag

install(suppress=[click])

TIMESTAMPER = structlog.processors.TimeStamper(fmt="iso")

LEVELED_TIMESTAMPED_PRE_CHAIN = frozenset(
    (
        # Add the log level and a timestamp to the event_dict if the log entry
        # is not from structlog.
        structlog.stdlib.add_log_level,
        TIMESTAMPER,
    )
)


def rich_exception_formatter_factory(
    color_system: str = "auto",
    no_color: bool | None = None,
) -> Callable[[TextIO, structlog.types.ExcInfo], None]:
    """Create an exception formatter for logging using the rich package.

    Examples:
    >>> rich_traceback = rich_exception_formatter_factory(color_system="truecolor")
    >>> plane_rich_traceback = rich_exception_formatter_factory(no_color=True)

    Args:
        color_system: The color system supported by your terminal.
        no_color: Enabled no color mode, or None to auto detect. Defaults to None.

    Returns:
        Exception formatter function.
    """

    def _traceback(sio, exc_info) -> None:
        sio.write("\n")
        Console(file=sio, color_system=color_system, no_color=no_color).print(
            Traceback.from_exception(*exc_info, show_locals=True)
        )

    return _traceback


def plain_rich_traceback(sio: TextIO, exc_info: structlog.types.ExcInfo) -> None:
    """Pretty-print `exc_info` to `sio` using the rich package, with colors disabled.

    To be passed into `ConsoleRenderer`'s `exception_formatter` argument.

    Args:
        sio: Return of open() in text mode.
        exc_info: Execution info.
    """
    sio.write("\n")
    Console(file=sio, no_color=True).print(
        Traceback.from_exception(*exc_info, show_locals=True)
    )


def _process_formatter(processor: Processor) -> structlog.stdlib.ProcessorFormatter:
    """Use _process_formatter to configure a structlog.stdlib.ProcessFormatter.

    It will automatically add log level and timestamp fields to any log entries not originating from structlog.

    Args:
        processor: A structlog message processor such as structlog.dev.ConsoleRenderer.

    Returns:
        A configured log processor.
    """
    return structlog.stdlib.ProcessorFormatter(
        processor=processor, foreign_pre_chain=LEVELED_TIMESTAMPED_PRE_CHAIN
    )


def console_log_formatter(colors: bool = False) -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter for console rendering that supports colorization.

    Args:
        colors: Add color to output.

    Returns:
        A configured console log formatter.
    """
    colors = colors and not get_no_color_flag()

    if colors:
        exception_formatter = rich_exception_formatter_factory(color_system="truecolor")
    else:
        exception_formatter = rich_exception_formatter_factory(no_color=True)

    return _process_formatter(
        structlog.dev.ConsoleRenderer(
            colors=colors, exception_formatter=exception_formatter
        )
    )


def key_value_formatter(
    sort_keys: bool = False,
    key_order: Sequence[str] | None = None,
    drop_missing: bool = False,
) -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter that renders lines in key=value format.

    Args:
        sort_keys: Whether to sort keys when formatting.
        key_order: List of keys that should be rendered in this exact order. Missing keys will be rendered as None, extra keys depending on *sort_keys* and the dict class.
        drop_missing: When True, extra keys in *key_order* will be dropped rather than rendered as None.

    Returns:
        A configured key=value formatter.
    """
    return _process_formatter(
        processor=structlog.processors.KeyValueRenderer(
            sort_keys=sort_keys, key_order=key_order, drop_missing=drop_missing
        )
    )


def json_formatter() -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter that renders lines in JSON format.

    Returns:
        A configured JSON formatter.
    """
    return _process_formatter(processor=structlog.processors.JSONRenderer())



"""add resource type to embed token

Revision ID: 23ea52e6d784
Revises: ceb00d7ff3bd
Create Date: 2020-02-12 09:29:31.592426

"""
import sqlalchemy as sa
import sqlalchemy.orm
from alembic import op

from meltano.migrations.utils.dialect_typing import (
    get_dialect_name,
    max_string_length_for_dialect,
)

# revision identifiers, used by Alembic.
revision = "23ea52e6d784"
down_revision = "ceb00d7ff3bd"
branch_labels = None
depends_on = None

Session = sa.orm.sessionmaker()


def upgrade():
    dialect_name = get_dialect_name()
    max_string_length = max_string_length_for_dialect(dialect_name)

    op.add_column(
        "embed_tokens", sa.Column("resource_type", sa.String(max_string_length))
    )

    metadata = sa.MetaData(bind=op.get_bind())
    Embed_Tokens = sa.Table("embed_tokens", metadata, autoload=True)
    op.execute(Embed_Tokens.update().values({"resource_type": "report"}))


def downgrade():
    op.drop_column("embed_tokens", "resource_type")



"""Fun Dragon CLI command."""
from __future__ import annotations

import random

import click

from meltano.core.cli_messages import (
    DRAGON_0,
    DRAGON_1,
    DRAGON_2,
    DRAGON_3,
    DRAGON_4,
    MELTY,
)

from . import cli


@cli.command(short_help="Summon a dragon!")
@click.pass_context
def dragon(ctx):
    """Summon a dragon."""
    dragon_list = [
        MELTY,
        DRAGON_0,
        DRAGON_1,
        DRAGON_2,
        DRAGON_3,
        DRAGON_4,
    ]
    click.secho(random.choice(dragon_list), nl=False)  # noqa: S311



"""Manage Python virtual environments."""

from __future__ import annotations

import asyncio
import hashlib
import logging
import os
import platform
import shutil
import subprocess
import sys
from asyncio.subprocess import Process
from collections import namedtuple
from collections.abc import Iterable
from pathlib import Path

from meltano.core.error import AsyncSubprocessError
from meltano.core.project import Project

logger = logging.getLogger(__name__)

VenvSpecs = namedtuple("VenvSpecs", ("lib_dir", "bin_dir", "site_packages_dir"))

POSIX = VenvSpecs(
    lib_dir="lib",
    bin_dir="bin",
    site_packages_dir=os.path.join(
        "lib",
        f"python{'.'.join(str(part) for part in sys.version_info[:2])}",
        "site-packages",
    ),
)

NT = VenvSpecs(
    lib_dir="Lib",
    bin_dir="Scripts",
    site_packages_dir=os.path.join("Lib", "site-packages"),
)

PLATFORM_SPECS = {"Linux": POSIX, "Darwin": POSIX, "Windows": NT}


def venv_platform_specs():
    """Get virtual environment sub-path info for the current platform.

    Raises:
        Exception: This platform is not supported.

    Returns:
        Virtual environment sub-path info for the current platform.
    """
    system = platform.system()
    try:
        return PLATFORM_SPECS[system]
    except KeyError as ex:
        raise Exception(f"Platform {system!r} not supported.") from ex


PIP_PACKAGES = ("pip", "setuptools==57.5.0", "wheel")


class VirtualEnv:
    """Info about a single virtual environment."""

    def __init__(self, root: Path):
        """Initialize the `VirtualEnv` instance.

        Args:
            root: The root directory of the virtual environment.
        """
        self.root = root.resolve()
        self.specs = venv_platform_specs()

    def __getattr__(self, key: str):
        """Get a specific attribute from this instance.

        Used to provide `VenvSpecs` attributes for this specific virtual environment.

        Args:
            key: The attribute name. Must be one of the `VenvSpecs` attributes.

        Returns:
            The root directory of this virtual environment joined to the requested
            platform-specific path using this platform's `VenvSpecs` instance.
        """
        return self.root / getattr(self.specs, key)

    def __str__(self):
        """_summary_.

        Returns:
            _description_.
        """
        return str(self.root)


async def exec_async(*args, **kwargs) -> Process:
    """Run an executable asyncronously in a subprocess.

    Args:
        args: Positional arguments for `asyncio.create_subprocess_exec`.
        kwargs: Keyword arguments for `asyncio.create_subprocess_exec`.

    Raises:
        AsyncSubprocessError: The command failed.

    Returns:
        The subprocess.
    """
    run = await asyncio.create_subprocess_exec(
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs,
    )
    await run.wait()

    if run.returncode != 0:
        raise AsyncSubprocessError("Command failed", run)

    return run


def fingerprint(pip_install_args: Iterable[str]) -> str:
    """Generate a hash identifying pip install args.

    Arguments are sorted and deduplicated before the hash is generated.

    Args:
        pip_install_args: Arguments for `pip install`.

    Returns:
        The SHA256 hash hex digest of the sorted set of pip install args.
    """
    return hashlib.sha256(" ".join(sorted(set(pip_install_args))).encode()).hexdigest()


class VenvService:  # noqa: WPS214
    """Manages virtual environments.

    The methods in this class are not threadsafe.
    """

    def __init__(self, project: Project, namespace: str = "", name: str = ""):
        """Initialize the `VenvService`.

        Args:
            project: The Meltano project.
            namespace: The namespace for the venv, e.g. a Plugin type.
            name: The name of the venv, e.g. a Plugin name.
        """
        self.project = project
        self.namespace = namespace
        self.name = name
        self.venv = VirtualEnv(self.project.venvs_dir(namespace, name))
        self.plugin_fingerprint_path = self.venv.root / ".meltano_plugin_fingerprint"

    async def install(self, pip_install_args: list[str], clean: bool = False) -> None:
        """Configure a virtual environment, then run pip install with the given args.

        Args:
            pip_install_args: Arguments passed to `pip install`.
            clean: Whether to not attempt to use an existing virtual environment.
        """
        if not clean and self.requires_clean_install(pip_install_args):
            logger.debug(
                f"Packages for '{self.namespace}/{self.name}' have changed so performing a clean install."
            )
            clean = True

        self.clean_run_files()
        await self._pip_install(pip_install_args=pip_install_args, clean=clean)
        self.write_fingerprint(pip_install_args)

    def requires_clean_install(self, pip_install_args: list[str]) -> bool:
        """Determine whether a clean install is needed.

        Args:
            pip_install_args: The arguments being passed to `pip install`, used
                for fingerprinting the installation.

        Returns:
            Whether virtual environment doesn't exist or can't be reused.
        """
        # A generator function is used to perform the checks lazily
        def checks():
            # The Python installation used to create this venv no longer exists
            yield not self.exec_path("python").exists()
            # The deprecated `meltano_venv.pth` feature is used by this venv
            yield self.venv.site_packages_dir.joinpath("meltano_venv.pth").exists()
            # The fingerprint of the venv does not match the pip install args
            existing_fingerprint = self.read_fingerprint()
            yield existing_fingerprint is None
            yield existing_fingerprint != fingerprint(pip_install_args)

        return any(checks())

    def clean_run_files(self) -> None:
        """Destroy cached configuration files, if they exist."""
        try:
            shutil.rmtree(self.project.run_dir(self.name, make_dirs=False))
        except FileNotFoundError:
            logger.debug("No cached configuration files to remove")

    def clean(self) -> None:
        """Destroy the virtual environment, if it exists."""
        try:
            shutil.rmtree(self.venv.root)
            logger.debug(
                "Removed old virtual environment for '%s/%s'",  # noqa: WPS323
                self.namespace,
                self.name,
            )
        except FileNotFoundError:
            # If the VirtualEnv has never been created before do nothing
            logger.debug("No old virtual environment to remove")

    async def create(self) -> Process:
        """Create a new virtual environment.

        Raises:
            AsyncSubprocessError: The virtual environment could not be created.

        Returns:
            The Python process creating the virtual environment.
        """
        logger.debug(f"Creating virtual environment for '{self.namespace}/{self.name}'")
        try:
            return await exec_async(sys.executable, "-m", "venv", str(self.venv))
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                f"Could not create the virtualenv for '{self.namespace}/{self.name}'",
                err.process,
            ) from err

    async def upgrade_pip(self) -> Process:
        """Upgrade the `pip` package to the latest version in the virtual environment.

        Raises:
            AsyncSubprocessError: Failed to upgrade pip to the latest version.

        Returns:
            The process running `pip install --upgrade ...`.
        """
        logger.debug(f"Upgrading pip for '{self.namespace}/{self.name}'")
        try:
            return await self._pip_install(["--upgrade", *PIP_PACKAGES])
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                "Failed to upgrade pip to the latest version.", err.process
            ) from err

    def read_fingerprint(self) -> str | None:
        """Get the fingerprint of the existing virtual environment.

        Returns:
            The fingerprint of the existing virtual environment if it exists.
            `None` otherwise.
        """
        if not self.plugin_fingerprint_path.exists():
            return None
        with open(self.plugin_fingerprint_path) as fingerprint_file:
            return fingerprint_file.read()

    def write_fingerprint(self, pip_install_args: list[str]) -> None:
        """Save the fingerprint for this installation.

        Args:
            pip_install_args: The arguments being passed to `pip install`.
        """
        with open(self.plugin_fingerprint_path, "wt") as fingerprint_file:
            fingerprint_file.write(fingerprint(pip_install_args))

    def exec_path(self, executable: str) -> Path:
        """Return the absolute path for the given executable in the virtual environment.

        Args:
            executable: The path to the executable relative to the venv bin directory.

        Returns:
            The venv bin directory joined to the provided executable.
        """
        absolute_executable = self.venv.bin_dir / executable
        if platform.system() != "Windows":
            return absolute_executable

        # On Windows, try using the '.exe' suffixed version if it exists. Use the
        # regular executable path as a fallback (and for backwards compatibility).
        absolute_executable_windows = absolute_executable.with_suffix(".exe")
        return (
            absolute_executable_windows
            if absolute_executable_windows.exists()
            else absolute_executable
        )

    async def _pip_install(
        self, pip_install_args: list[str], clean: bool = False
    ) -> Process:
        """Install a package using `pip` in the proper virtual environment.

        Args:
            pip_install_args: The arguments to pass to `pip install`.
            clean: Whether the installation should be done in a clean venv.

        Raises:
            AsyncSubprocessError: The command failed.

        Returns:
            The process running `pip install` with the provided args.
        """
        if clean:
            self.clean()
            await self.create()
            await self.upgrade_pip()

        pip_install_args_str = " ".join(pip_install_args)
        log_msg_prefix = (
            f"Upgrading with args {pip_install_args_str!r} in existing"
            if "--upgrade" in pip_install_args
            else f"Installing with args {pip_install_args_str!r} into"
        )
        logger.debug(
            f"{log_msg_prefix} virtual environment for '{self.namespace}/{self.name}'"
        )

        try:
            return await exec_async(
                str(self.exec_path("python")), "-m", "pip", "install", *pip_install_args
            )
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                f"Failed to install plugin '{self.name}'.", err.process
            ) from err



"""Exception context for the Snowplow tracker."""

from __future__ import annotations

import sys
import uuid
from pathlib import Path
from types import TracebackType
from typing import Dict, List, Union

from snowplow_tracker import SelfDescribingJson

from meltano.core.tracking.schemas import ExceptionContextSchema
from meltano.core.utils import hash_sha256

BASE_PATHS = (sys.prefix, sys.exec_prefix, sys.base_prefix, sys.base_exec_prefix)

TracebackLevelsJSON = List[Dict[str, Union[str, int]]]
ExceptionContextJSON = Dict[
    str, Union[str, TracebackLevelsJSON, "ExceptionContextJSON"]
]


class ExceptionContext(SelfDescribingJson):
    """Exception context for the Snowplow tracker."""

    def __init__(self):
        """Initialize the exceptions context with the exceptions currently being handled."""
        ex = sys.exc_info()[1]
        super().__init__(
            ExceptionContextSchema.url,
            {
                "context_uuid": str(uuid.uuid4()),
                "exception": None if ex is None else get_exception_json(ex),
            },
        )


def get_exception_json(ex: BaseException) -> ExceptionContextJSON:
    """Get anonymized telemetry data detailing an exception.

    Args:
        ex: The exception from which data will be extracted.

    Returns:
        A JSON-compatible dictionary of anonymized telemetry data compliant with the exception
        context schema for an exception.
    """
    cause, context, tb = ex.__cause__, ex.__context__, ex.__traceback__  # noqa: WPS609
    return {
        "type": str(type(ex).__qualname__),
        "str_hash": hash_sha256(str(ex)),
        "repr_hash": hash_sha256(repr(ex)),
        "traceback": None if tb is None else get_traceback_json(tb),
        "cause": None if cause is None else get_exception_json(cause),
        "context": None if context is None else get_exception_json(context),
    }


def get_traceback_json(tb: TracebackType) -> TracebackLevelsJSON:
    """Get anonymized telemetry data detailing a traceback.

    Args:
        tb: The traceback from which data will be extracted.

    Returns:
        A list where each successive element details a lower level in the traceback.
    """
    levels = []
    while tb:
        levels.append(
            {
                "file": get_relative_traceback_path(tb),
                "line_number": tb.tb_lineno,
            }
        )
        tb = tb.tb_next
    return levels


def get_relative_traceback_path(tb: TracebackType) -> str | None:
    """Get an anonymous path from a traceback by making it relative if possible.

    The path is made relative to the first element in `BASE_PATHS` it can be made relative to.

    Args:
        tb: The traceback from which to extract the path info.

    Returns:
        The first valid option of the following is returned:
        - If the path is `'<stdin>'`: `'<stdin>'`
        - If possible: the path made relative to a path in `BASE_PATHS`
        - If the file name is `__init__.py` or `__main__.py`: `'.../<module name>/<file name>.py'`.
        - Otherwise: `'.../<filename>.py'`.
    """
    try:
        str_path = tb.tb_frame.f_code.co_filename
    except Exception:
        return None

    if str_path == "<stdin>":
        return str_path

    path = Path(str_path)

    for base_path in BASE_PATHS:
        try:
            return path.relative_to(base_path).as_posix()
        except ValueError:  # Path could not be made relative to `base_path`
            pass  # Try making it relative to the next base path

    if path.parts[-1] in {"__init__.py", "__main__.py"}:
        # Include the module directory if the file is `__init__.py` or `__main__.py`
        return f".../{path.parts[-2]}/{path.parts[-1]}"
    return f".../{path.parts[-1]}"



"""Service for managing meltano schedules."""

from __future__ import annotations

import logging
import subprocess
from datetime import date, datetime

from croniter import croniter

from meltano.core.error import MeltanoError
from meltano.core.setting_definition import SettingMissingError

from .meltano_invoker import MeltanoInvoker
from .plugin import PluginType
from .plugin.settings_service import PluginSettingsService
from .plugin_discovery_service import PluginNotFoundError
from .project import Project
from .project_plugins_service import ProjectPluginsService
from .schedule import Schedule
from .task_sets_service import TaskSetsService
from .utils import NotFound, coerce_datetime, find_named, iso8601_datetime


class ScheduleAlreadyExistsError(MeltanoError):
    """Occurs when a schedule already exists."""

    def __init__(self, schedule: Schedule):
        """Initialize the exception.

        Args:
            schedule: The schedule that already exists.
        """
        self.schedule = schedule
        super().__init__(reason=f"Schedule '{self.schedule.name}' already exists")


class ScheduleDoesNotExistError(MeltanoError):
    """Occurs when a schedule does not exist."""

    def __init__(self, name: str):
        """Initialize the exception.

        Args:
            name: The name of the schedule that does not exist.

        """
        self.name = name

        super().__init__(
            reason=f"Schedule '{name}' does not exist",
            instruction="Use `meltano schedule add` to add a schedule",
        )


class ScheduleNotFoundError(MeltanoError):
    """Occurs when a schedule for a namespace cannot be found."""

    def __init__(self, namespace: str):
        """Initialize the exception.

        Args:
            namespace: The namespace that had no associated schedules.
        """
        self.namespace = namespace

        reason = f"No schedule found for namespace {self.namespace}"
        instruction = "Use `meltano schedule add` to add a schedule"
        super().__init__(reason, instruction)


class BadCronError(MeltanoError):
    """Occurs when a cron expression is invalid."""

    def __init__(self, cron: str):
        """Initialize the exception.

        Args:
            cron: The cron expression that is invalid.
        """
        self.cron = cron

        reason = f"Invalid Cron expression or alias: '{self.cron}'"
        instruction = "Please use a valid cron expression"
        super().__init__(reason, instruction)


class ScheduleService:  # noqa: WPS214
    """Service for managing schedules."""

    def __init__(
        self, project: Project, plugins_service: ProjectPluginsService | None = None
    ):
        """Initialize a ScheduleService for a project to manage a projects schedules.

        Args:
            project: The project whos schedules you wish to interact with.
            plugins_service: The project plugins service.
        """
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)
        self.task_sets_service = TaskSetsService(project)

    def add_elt(
        self,
        session,
        name,
        extractor: str,
        loader: str,
        transform: str,
        interval: str,
        start_date: datetime | None = None,
        **env,
    ) -> Schedule:
        """Add a scheduled legacy elt task.

        Args:
            session: The active sqlalchemy session.
            name: The name of the schedule.
            extractor:  The name of the extractor.
            loader: The name of the loader.
            transform: The transform statement (eg: skip, only, run)
            interval: The interval of the elt job.
            start_date: The start date of the elt job.
            env: The env for this scheduled elt job.

        Returns:
            The added schedule.
        """
        start_date = coerce_datetime(start_date) or self.default_start_date(  # TODO
            session, extractor
        )

        schedule = Schedule(
            name, extractor, loader, transform, interval, start_date, env=env
        )
        return self.add_schedule(schedule)

    def add(self, name: str, job: str, interval: str, **env) -> Schedule:
        """Add a scheduled job.

        Args:
            name: The name of the schedule.
            job: The name of the job.
            interval: The interval of the job.
            env: The env for this scheduled job

        Returns:
            The added schedule.
        """
        schedule = Schedule(name=name, job=job, interval=interval, env=env)
        self.add_schedule(schedule)
        return schedule

    def remove(self, name) -> str:
        """Remove a schedule from the project.

        Args:
            name: The name of the schedule.

        Returns:
            The name of the removed schedule.
        """
        return self.remove_schedule(name)

    def default_start_date(self, session, extractor: str) -> datetime:
        """Obtain the default start date for an elt schedule.

        Args:
            session: The session to use.
            extractor: The extractor to get the start_date from.

        Returns:
            The start_date of the extractor, or now.
        """
        extractor_plugin = self.plugins_service.find_plugin(
            extractor, plugin_type=PluginType.EXTRACTORS
        )
        start_date: str | datetime | date | None = None
        try:
            settings_service = PluginSettingsService(
                self.project, extractor_plugin, plugins_service=self.plugins_service
            )
            start_date = settings_service.get("start_date", session=session)
        except SettingMissingError:
            logging.debug(f"`start_date` not found in {extractor_plugin}.")

        # TODO: this coercion should be handled by the `kind` attribute
        # on the actual setting
        if isinstance(start_date, date):
            return coerce_datetime(start_date)

        if isinstance(start_date, datetime):
            return start_date

        return iso8601_datetime(start_date) or datetime.utcnow()

    def add_schedule(self, schedule: Schedule) -> Schedule:
        """Add a schedule to the project.

        Args:
            schedule: The schedule to add.

        Returns:
            The added schedule.

        Raises:
            BadCronError: If the cron expression is invalid.
            ScheduleAlreadyExistsError: If a schedule with the same name already exists.
        """
        if (
            schedule.interval is not None
            and schedule.interval != "@once"
            and not croniter.is_valid(schedule.interval)
        ):
            raise BadCronError(schedule.interval)

        with self.project.meltano_update() as meltano:
            # guard if it already exists
            if schedule in meltano.schedules:
                raise ScheduleAlreadyExistsError(schedule)

            meltano.schedules.append(schedule)

        return schedule

    def remove_schedule(self, name: str) -> str:
        """Remove a schedule.

        Args:
            name: The name of the schedule.

        Returns:
            The name of the schedule.

        Raises:
            ScheduleDoesNotExistError: If the schedule does not exist.
        """
        with self.project.meltano_update() as meltano:
            try:
                # guard if it doesn't exist
                schedule = find_named(self.schedules(), name)
            except NotFound:
                raise ScheduleDoesNotExistError(name)

            # find the schedules plugin config
            meltano.schedules.remove(schedule)

        return name

    def update_schedule(self, schedule: Schedule):
        """Update a schedule.

        Args:
            schedule: The schedule to update.

        Raises:
            ScheduleDoesNotExistError: If the schedule doesn't exist.
        """
        with self.project.meltano_update() as meltano:
            try:
                idx = meltano.schedules.index(schedule)
                meltano.schedules[idx] = schedule
            except ValueError:
                raise ScheduleDoesNotExistError(schedule.name)

    def find_namespace_schedule(self, namespace: str) -> Schedule:
        """Search for a Schedule that runs for a certain plugin namespace.

        Example:
            `tap_carbon` would yield the first schedule that runs for the `tap-carbon` extractor.

        Args:
            namespace: The plugin namespace to search.

        Returns:
            The schedule

        Raises:
            ScheduleNotFoundError: If no schedule is found.
        """
        try:
            extractor = self.plugins_service.find_plugin_by_namespace(
                PluginType.EXTRACTORS, namespace
            )

            return next(
                schedule
                for schedule in self.schedules()
                if not schedule.job and schedule.extractor == extractor.name
            )
        except (PluginNotFoundError, StopIteration) as err:
            raise ScheduleNotFoundError(namespace) from err

    def schedules(self) -> list[Schedule]:
        """Return all schedules in the project.

        Returns:
            The list of schedules.
        """
        return self.project.meltano.schedules.copy()

    def find_schedule(self, name) -> Schedule:
        """Find a schedule by name.

        Args:
            name: the name of the schedule to find

        Returns:
            Schedule: the schedule with the given name

        Raises:
            ScheduleNotFoundError: if the schedule does not exist
        """
        try:
            return find_named(self.schedules(), name)
        except StopIteration as err:
            raise ScheduleNotFoundError(name) from err

    def run(
        self, schedule: Schedule, *args, env: dict | None = None, **kwargs
    ) -> subprocess.CompletedProcess:
        """Run a scheduled elt task or named job.

        Args:
            schedule: The schedule whos elt task or named job to run.
            args: The arguments to pass to the elt task or named job.
            env: The environment to run the elt task or named job in.
            kwargs: The keyword arguments to pass to the elt task or named job.

        Returns:
            The completed process.
        """
        if env is None:
            env = {}

        if schedule.job:
            return MeltanoInvoker(self.project).invoke(
                ["run", *args, schedule.job],
                env={**schedule.env, **env},
                **kwargs,
            )

        return MeltanoInvoker(self.project).invoke(
            ["elt", *schedule.elt_args, *args], env={**schedule.env, **env}, **kwargs
        )



from __future__ import annotations

import asyncio
import logging
import subprocess
import sys
from contextlib import suppress

from meltano.core.elt_context import ELTContext
from meltano.core.logging import capture_subprocess_output
from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import human_size

from . import Runner, RunnerError


class SingerRunner(Runner):
    def __init__(self, elt_context: ELTContext):
        self.context = elt_context

        self.project_settings_service = ProjectSettingsService(
            self.context.project,
            config_service=self.context.plugins_service.config_service,
        )

    def stop(self, process, **wait_args):
        while True:
            try:
                code = process.wait(**wait_args)
                logging.debug(f"{process} exited with {code}")
                return code
            except subprocess.TimeoutExpired:
                process.kill()
                logging.error(f"{process} was killed.")

    async def invoke(  # noqa: WPS210, WPS213, WPS217, WPS231, WPS238
        self,
        tap: PluginInvoker,
        target: PluginInvoker,
        extractor_log=None,
        loader_log=None,
        extractor_out=None,
        loader_out=None,
    ):
        """Invoke tap and target together."""
        extractor_log = extractor_log or sys.stderr
        loader_log = loader_log or sys.stderr

        # The StreamReader line length limit also acts as half the buffer size,
        # which cannot be set directly:
        # - https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L395-L396
        # - https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L482
        stream_buffer_size = self.project_settings_service.get("elt.buffer_size")
        line_length_limit = stream_buffer_size // 2

        # Start tap
        try:
            p_tap = await tap.invoke_async(
                limit=line_length_limit,
                stdout=asyncio.subprocess.PIPE,  # Singer messages
                stderr=asyncio.subprocess.PIPE,  # Log
            )
        except Exception as err:
            raise RunnerError(f"Cannot start extractor: {err}") from err

        # Start target
        try:
            p_target = await target.invoke_async(
                limit=line_length_limit,
                stdin=asyncio.subprocess.PIPE,  # Singer messages
                stdout=asyncio.subprocess.PIPE,  # Singer state
                stderr=asyncio.subprocess.PIPE,  # Log
            )
        except Exception as err:
            raise RunnerError(f"Cannot start loader: {err}") from err

        # Process tap output
        tap_outputs = [p_target.stdin]
        if extractor_out:
            tap_outputs.insert(0, extractor_out)

        tap_stdout_future = asyncio.ensure_future(
            # forward subproc stdout to tap_outputs (i.e. targets stdin)
            capture_subprocess_output(p_tap.stdout, *tap_outputs)
        )
        tap_stderr_future = asyncio.ensure_future(
            capture_subprocess_output(p_tap.stderr, extractor_log)
        )

        # Process target output
        target_outputs = []
        if target.output_handlers:
            target_outputs = target.output_handlers.get(target.StdioSource.STDOUT, [])

        if loader_out:
            target_outputs.insert(0, loader_out)

        target_stdout_future = asyncio.ensure_future(
            capture_subprocess_output(p_target.stdout, *target_outputs)
        )
        target_stderr_future = asyncio.ensure_future(
            capture_subprocess_output(p_target.stderr, loader_log)
        )

        # Wait for tap or target to complete, or for one of the output handlers to raise an exception.
        tap_process_future = asyncio.ensure_future(p_tap.wait())
        target_process_future = asyncio.ensure_future(p_target.wait())
        output_exception_future = asyncio.ensure_future(
            asyncio.wait(
                [
                    tap_stdout_future,
                    tap_stderr_future,
                    target_stdout_future,
                    target_stderr_future,
                ],
                return_when=asyncio.FIRST_EXCEPTION,
            ),
        )

        done, _ = await asyncio.wait(
            [tap_process_future, target_process_future, output_exception_future],
            return_when=asyncio.FIRST_COMPLETED,
        )

        # If `output_exception_future` completes first, one of the output handlers raised an exception or all completed successfully.
        if output_exception_future in done:
            output_futures_done, _ = output_exception_future.result()
            output_futures_failed = [
                future
                for future in output_futures_done
                if future.exception() is not None
            ]

            if output_futures_failed:
                # If any output handler raised an exception, re-raise it.

                # Special behavior for the tap stdout handler raising a line length limit error.
                if tap_stdout_future in output_futures_failed:
                    self._handle_tap_line_length_limit_error(
                        tap_stdout_future.exception(),
                        line_length_limit=line_length_limit,
                        stream_buffer_size=stream_buffer_size,
                    )

                failed_future = output_futures_failed.pop()
                raise failed_future.exception()
            else:
                # If all of the output handlers completed without raising an exception,
                # we still need to wait for the tap or target to complete.
                done, _ = await asyncio.wait(
                    [tap_process_future, target_process_future],
                    return_when=asyncio.FIRST_COMPLETED,
                )

        if target_process_future in done:
            target_code = target_process_future.result()

            if tap_process_future in done:
                tap_code = tap_process_future.result()
            else:
                # If the target completes before the tap, it failed before processing all tap output

                # Kill tap and cancel output processing since there's no more target to forward messages to
                p_tap.kill()
                await tap_process_future
                tap_stdout_future.cancel()
                tap_stderr_future.cancel()

                # Pretend the tap finished successfully since it didn't itself fail
                tap_code = 0

            # Wait for all buffered target output to be processed
            await asyncio.wait([target_stdout_future, target_stderr_future])
        else:  # if tap_process_future in done:
            # If the tap completes before the target, the target should have a chance to process all tap output
            tap_code = tap_process_future.result()

            # Wait for all buffered tap output to be processed
            await asyncio.wait([tap_stdout_future, tap_stderr_future])

            # Close target stdin so process can complete naturally
            p_target.stdin.close()
            with suppress(AttributeError):  # `wait_closed` is Python 3.7+
                await p_target.stdin.wait_closed()

            # Wait for all buffered target output to be processed
            await asyncio.wait([target_stdout_future, target_stderr_future])

            # Wait for target to complete
            target_code = await target_process_future

        if tap_code and target_code:
            raise RunnerError(
                "Extractor and loader failed",
                {PluginType.EXTRACTORS: tap_code, PluginType.LOADERS: target_code},
            )
        elif tap_code:
            raise RunnerError("Extractor failed", {PluginType.EXTRACTORS: tap_code})
        elif target_code:
            raise RunnerError("Loader failed", {PluginType.LOADERS: target_code})

    def dry_run(self, tap: PluginInvoker, target: PluginInvoker):
        logging.info("Dry run:")
        logging.info(f"\textractor: {tap.plugin.name} at '{tap.exec_path()}'")
        logging.info(f"\tloader: {target.plugin.name} at '{target.exec_path()}'")

    async def run(
        self, extractor_log=None, loader_log=None, extractor_out=None, loader_out=None
    ):
        tap = self.context.extractor_invoker()
        target = self.context.loader_invoker()

        if self.context.dry_run:
            return self.dry_run(tap, target)

        async with tap.prepared(self.context.session):
            async with target.prepared(self.context.session):
                await self.invoke(
                    tap,
                    target,
                    extractor_log=extractor_log,
                    loader_log=loader_log,
                    extractor_out=extractor_out,
                    loader_out=loader_out,
                )

    def _handle_tap_line_length_limit_error(
        self,
        exception,
        line_length_limit,
        stream_buffer_size,
    ):
        # StreamReader.readline can raise a ValueError wrapping a LimitOverrunError:
        # https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L549
        if not isinstance(exception, ValueError):
            return

        exception = exception.__context__  # noqa: WPS609
        if not isinstance(exception, asyncio.LimitOverrunError):
            return

        logging.error(
            f"The extractor generated a message exceeding the message size limit of {human_size(line_length_limit)} (half the buffer size of {human_size(stream_buffer_size)})."
        )
        logging.error(
            "To let this message be processed, increase the 'elt.buffer_size' setting to at least double the size of the largest expected message, and try again."
        )
        logging.error(
            "To learn more, visit https://docs.meltano.com/reference/settings#eltbuffer_size"
        )
        raise RunnerError("Output line length limit exceeded") from exception



"""Add run_id to Job

Revision ID: 53e97221d99f
Revises: 6ef30ab7b8e5
Create Date: 2019-10-10 13:12:55.147164

"""
import uuid

import sqlalchemy as sa
import sqlalchemy.orm
import sqlalchemy.types as types
from alembic import op

from meltano.migrations import GUID

# revision identifiers, used by Alembic.
revision = "53e97221d99f"
down_revision = "6ef30ab7b8e5"
branch_labels = None
depends_on = None

Session = sa.orm.sessionmaker()


def upgrade():
    op.add_column("job", sa.Column("run_id", GUID))

    metadata = sa.MetaData(bind=op.get_bind())
    session = Session(bind=op.get_bind())
    Job = sa.Table("job", metadata, autoload=True)

    for job in session.query(Job):
        job.run_id = uuid.uuid4()

    session.commit()


def downgrade():
    op.drop_column("job", "run_id")



"""Meltano Projects."""


from __future__ import annotations

import errno
import logging
import os
import sys
import threading
from contextlib import contextmanager, suppress
from pathlib import Path
from typing import TYPE_CHECKING, Any

import fasteners
from cached_property import cached_property
from dotenv import dotenv_values
from werkzeug.utils import secure_filename

from meltano.core import yaml
from meltano.core.behavior.versioned import Versioned
from meltano.core.environment import Environment
from meltano.core.error import EmptyMeltanoFileException, Error
from meltano.core.plugin.base import PluginRef
from meltano.core.project_files import ProjectFiles
from meltano.core.utils import makedirs, truthy

if TYPE_CHECKING:
    from .meltano_file import MeltanoFile as MeltanoFileTypeHint


logger = logging.getLogger(__name__)


PROJECT_ROOT_ENV = "MELTANO_PROJECT_ROOT"
PROJECT_ENVIRONMENT_ENV = "MELTANO_ENVIRONMENT"
PROJECT_READONLY_ENV = "MELTANO_PROJECT_READONLY"
PROJECT_SYS_DIR_ROOT = "MELTANO_SYS_DIR_ROOT"


class ProjectNotFound(Error):
    """Occurs when a Project is instantiated outside of a meltano project structure."""

    def __init__(self, project: Project):
        """Instantiate the error.

        Args:
            project: the name of the project which cannot be found
        """
        super().__init__(
            f"Cannot find `{project.meltanofile}`. Are you in a meltano project?"
        )


class ProjectReadonly(Error):
    """Occurs when attempting to update a readonly project."""

    def __init__(self):
        """Instantiate the error."""
        super().__init__("This Meltano project is deployed as read-only")


def walk_parent_directories():
    """Yield each directory starting with the current up to the root.

    Yields:
        parent directories
    """
    directory = os.getcwd()
    while True:
        yield directory

        parent_directory = os.path.dirname(directory)
        if parent_directory == directory:
            return
        directory = parent_directory


class Project(Versioned):  # noqa: WPS214
    """Represent the current Meltano project from a file-system perspective."""

    __version__ = 1
    _activate_lock = threading.Lock()
    _find_lock = threading.Lock()
    _meltano_rw_lock = fasteners.ReaderWriterLock()
    _default = None

    def __init__(self, root: os.PathLike):
        """Instantiate a Project from its root directory.

        Args:
            root: the root directory for the project
        """
        self.root = Path(root).resolve()
        self.sys_dir_root = Path(
            os.getenv(PROJECT_SYS_DIR_ROOT, self.root / ".meltano")
        ).resolve()
        self.readonly = False
        self.active_environment: Environment | None = None

    @cached_property
    def _meltano_interprocess_lock(self):
        return fasteners.InterProcessLock(self.run_dir("meltano.yml.lock"))

    @property
    def env(self):
        """Get environment variables for this project.

        Returns:
            dict of environment variables and values for this project.
        """
        environment_name = (
            self.active_environment.name if self.active_environment else ""
        )
        return {
            PROJECT_ROOT_ENV: str(self.root),
            PROJECT_ENVIRONMENT_ENV: environment_name,
            PROJECT_SYS_DIR_ROOT: str(self.sys_dir_root),
        }

    @classmethod
    @fasteners.locked(lock="_activate_lock")
    def activate(cls, project: Project):
        """Activate the given Project.

        Args:
            project: the Project to activate

        Raises:
            OSError: if project cannot be activated due to unsupported OS
        """
        import ctypes

        project.ensure_compatible()

        # create a symlink to our current binary
        try:
            # check if running on Windows
            if os.name == "nt":
                executable = Path(sys.executable).parent / "meltano.exe"
                # Admin privileges are required to create symlinks on Windows
                if ctypes.windll.shell32.IsUserAnAdmin():
                    if executable.is_file():
                        project.run_dir().joinpath("bin").symlink_to(executable)
                    else:
                        logger.warn(
                            "Could not create symlink: meltano.exe not "
                            f"present in {str(Path(sys.executable).parent)}"
                        )
                else:
                    logger.warn(
                        "Failed to create symlink to 'meltano.exe': "
                        "administrator privilege required"
                    )
            else:
                executable = Path(sys.executable).parent / "meltano"
                if executable.is_file():
                    project.run_dir().joinpath("bin").symlink_to(executable)
        except FileExistsError:
            pass
        except OSError as error:
            if error.errno == errno.EOPNOTSUPP:
                logger.warning(
                    f"Could not create symlink: {error}\nPlease make sure that the underlying filesystem supports symlinks."
                )
            else:
                raise

        logger.debug(f"Activated project at {project.root}")

        # set the default project
        cls._default = project

    @classmethod
    def deactivate(cls):
        """Deactivate the given Project."""
        cls._default = None

    @property
    def file_version(self):
        """Get the version of Meltano found in this project's meltano.yml.

        Returns:
            the Project's meltano version
        """
        return self.meltano.version

    @classmethod
    @fasteners.locked(lock="_find_lock")
    def find(cls, project_root: Path | str | None = None, activate=True):
        """Find a Project.

        Args:
            project_root: The path to the root directory of the project. If not supplied,
                infer from PROJECT_ROOT_ENV or the current working directory and it's parents.
            activate: Save the found project so that future calls to `find` will
                continue to use this project.

        Returns:
            the found project

        Raises:
            ProjectNotFound: if the provided `project_root` is not a Meltano project, or
                the current working directory is not a Meltano project or a subfolder of one.
        """
        if cls._default:
            return cls._default

        project_root = project_root or os.getenv(PROJECT_ROOT_ENV)
        if project_root:
            project = Project(project_root)
            if not project.meltanofile.exists():
                raise ProjectNotFound(project)
        else:
            for directory in walk_parent_directories():
                project = Project(directory)
                if project.meltanofile.exists():
                    break
            if not project.meltanofile.exists():
                raise ProjectNotFound(Project(os.getcwd()))

        # if we activate a project using `find()`, it should
        # be set as the default project for future `find()`
        if activate:
            cls.activate(project)

        if truthy(os.getenv(PROJECT_READONLY_ENV, "false")):
            project.readonly = True

        return project

    @cached_property
    def project_files(self) -> ProjectFiles:
        """Return a singleton `ProjectFiles` file manager instance.

        Returns:
            `ProjectFiles` file manager.
        """
        return ProjectFiles(root=self.root, meltano_file_path=self.meltanofile)

    def clear_cache(self) -> None:
        """Clear cached project files (e.g. `meltano.yml`).

        This can be useful if the cached `ProjectFiles` object has been
        modified in-place, but not updated on-disk, and you need the on-disk
        version.
        """
        with suppress(KeyError):
            del self.__dict__["project_files"]

    @property
    def meltano(self) -> MeltanoFileTypeHint:
        """Return a copy of the current meltano config.

        Raises:
            EmptyMeltanoFileException: The `meltano.yml` file is empty.

        Returns:
            The current meltano config.
        """
        from meltano.core.meltano_file import MeltanoFile
        from meltano.core.settings_service import FEATURE_FLAG_PREFIX, FeatureFlags

        conf: dict[str, Any] = yaml.load(self.meltanofile)
        if conf is None:
            raise EmptyMeltanoFileException()

        lock = (
            self._meltano_rw_lock.write_lock
            if conf.get(f"{FEATURE_FLAG_PREFIX}.{FeatureFlags.ENABLE_UVICORN}", False)
            else self._meltano_rw_lock.read_lock
        )
        with lock():
            return MeltanoFile.parse(self.project_files.load())

    @contextmanager
    def meltano_update(self):
        """Yield the current meltano configuration.

        Update the meltanofile if the context ends gracefully.

        Yields:
            the current meltano configuration

        Raises:
            ProjectReadonly: This project is readonly.
            Exception: The project files could not be updated.
        """
        if self.readonly:
            raise ProjectReadonly

        from meltano.core.meltano_file import MeltanoFile

        with self._meltano_rw_lock.write_lock(), self._meltano_interprocess_lock:

            meltano_config = MeltanoFile.parse(self.project_files.load())
            yield meltano_config

            try:
                self.project_files.update(meltano_config.canonical())
            except Exception as err:
                logger.critical("Could not update meltano.yml: %s", err)  # noqa: WPS323
                raise

    def root_dir(self, *joinpaths):
        """Return the root directory of this project, optionally joined with path.

        Args:
            joinpaths: list of subdirs and/or file to join to project root.

        Returns:
            project root joined with provided subdirs and/or file
        """
        return self.root.joinpath(*joinpaths)

    @contextmanager
    def file_update(self):
        """Raise error if project is readonly.

        Used in context where project files would be updated.

        Yields:
            the project root

        Raises:
            ProjectReadonly: if the project is readonly
        """
        if self.readonly:
            raise ProjectReadonly

        yield self.root

    @property
    def meltanofile(self):
        """Get the path to this project's meltano.yml.

        Returns:
            the path to this project meltano.yml
        """
        return self.root.joinpath("meltano.yml")

    @property
    def dotenv(self):
        """Get the path to this project's .env file.

        Returns:
            the path to this project's .env file
        """
        return self.root.joinpath(".env")

    @property
    def dotenv_env(self):
        """Get values from this project's .env file.

        Returns:
            values found in this project's .env file
        """
        return dotenv_values(self.dotenv)

    def activate_environment(self, name: str) -> None:
        """Retrieve an environment configuration.

        Args:
            name: Name of the environment.
        """
        self.active_environment = Environment.find(self.meltano.environments, name)
        logger.info(f"Environment {name!r} is active")

    def deactivate_environment(self) -> None:
        """Deactivate the currently active environment."""
        self.active_environment = None

    @contextmanager
    def dotenv_update(self):
        """Raise error if project is readonly.

        Used in context where .env files would be updated.

        Yields:
            the .env file

        Raises:
            ProjectReadonly: if the project is readonly
        """
        if self.readonly:
            raise ProjectReadonly

        yield self.dotenv

    @makedirs
    def meltano_dir(self, *joinpaths, make_dirs: bool = True):
        """Path to the project `.meltano` directory.

        Args:
            joinpaths: Paths to join to the `.meltano` directory.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to `.meltano` dir optionally joined to given paths.
        """
        return self.sys_dir_root.joinpath(*joinpaths)

    @makedirs
    def analyze_dir(self, *joinpaths, make_dirs: bool = True):
        """Path to the project `analyze` directory.

        Args:
            joinpaths: Paths to join to the `analyze` directory.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to `analyze` dir optionally joined to given paths.
        """
        return self.root_dir("analyze", *joinpaths)

    @makedirs
    def extract_dir(self, *joinpaths, make_dirs: bool = True):
        """Path to the project `extract` directory.

        Args:
            joinpaths: Paths to join to the `extract` directory.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to `extract` dir optionally joined to given paths.
        """
        return self.root_dir("extract", *joinpaths)

    @makedirs
    def venvs_dir(self, *prefixes, make_dirs: bool = True):
        """Path to a `venv` directory in `.meltano`.

        Args:
            prefixes: Paths to prepend to the `venv` directory in `.meltano`.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to `venv` dir optionally prepended with given prefixes.
        """
        return self.meltano_dir(*prefixes, "venv", make_dirs=make_dirs)

    @makedirs
    def run_dir(self, *joinpaths, make_dirs: bool = True):
        """Path to the `run` directory in `.meltano`.

        Args:
            joinpaths: Paths to join to the `run` directory in `.meltano`.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to `run` dir optionally joined to given paths.
        """
        return self.meltano_dir("run", *joinpaths, make_dirs=make_dirs)

    @makedirs
    def logs_dir(self, *joinpaths, make_dirs: bool = True):
        """Path to the `logs` directory in `.meltano`.

        Args:
            joinpaths: Paths to join to the `logs` directory in `.meltano`.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to `logs` dir optionally joined to given paths.
        """
        return self.meltano_dir("logs", *joinpaths, make_dirs=make_dirs)

    @makedirs
    def job_dir(self, state_id, *joinpaths, make_dirs: bool = True):
        """Path to the `elt` directory in `.meltano/run`.

        Args:
            state_id: State ID of `run` dir.
            joinpaths: Paths to join to the `elt` directory in `.meltano`.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to `elt` dir optionally joined to given paths.
        """
        return self.run_dir(
            "elt", secure_filename(state_id), *joinpaths, make_dirs=make_dirs
        )

    @makedirs
    def job_logs_dir(self, state_id, *joinpaths, make_dirs: bool = True):
        """Path to the `elt` directory in `.meltano/logs`.

        Args:
            state_id: State ID of `logs` dir.
            joinpaths: Paths to join to the `elt` directory in `.meltano/logs`.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to `elt` dir optionally joined to given paths.
        """
        return self.logs_dir(
            "elt", secure_filename(state_id), *joinpaths, make_dirs=make_dirs
        )

    @makedirs
    def plugin_dir(self, plugin: PluginRef, *joinpaths, make_dirs: bool = True):
        """Path to the plugin installation directory in `.meltano`.

        Args:
            plugin: Plugin to retrieve or create directory for.
            joinpaths: Paths to join to the plugin installation directory in `.meltano`.
            make_dirs: Flag to make directories if not exists.

        Returns:
            Resolved path to plugin installation dir optionally joined to given paths.
        """
        return self.meltano_dir(
            plugin.type, plugin.name, *joinpaths, make_dirs=make_dirs
        )

    @makedirs
    def root_plugins_dir(self, *joinpaths: str, make_dirs: bool = True):
        """Path to the project `plugins` directory.

        Args:
            joinpaths: Paths to join with the project `plugins` directory.
            make_dirs: If True, create the directory hierarchy if it does not exist.

        Returns:
            Path to the project `plugins` directory.
        """
        return self.root_dir("plugins", *joinpaths)

    @makedirs
    def plugin_lock_path(
        self,
        plugin_type: str,
        plugin_name: str,
        variant_name: str | None = None,
        make_dirs: bool = True,
    ):
        """Path to the project lock file.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant_name: The plugin variant name.
            make_dirs: If True, create the directory hierarchy if it does not exist.

        Returns:
            Path to the plugin lock file.
        """
        filename = f"{plugin_name}"

        if variant_name:
            filename = f"{filename}--{variant_name}"

        return self.root_plugins_dir(
            plugin_type,
            f"{filename}.lock",
            make_dirs=make_dirs,
        )

    def __eq__(self, other):
        """Project equivalence check.

        Args:
            other: The other Project instance to check against.

        Returns:
            True if Projects are equal.
        """
        return hasattr(other, "root") and self.root == other.root  # noqa: WPS421

    def __hash__(self):
        """Project hash.

        Returns:
            Project hash.
        """
        return self.root.__hash__()  # noqa: WPS609



"""Rename job to job_run and job_id to job_name

Revision ID: 5b43800443d1
Revises: 13e8639c6d2b
Create Date: 2022-06-29 14:28:51.673195

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5b43800443d1"
down_revision = "13e8639c6d2b"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("job", "job_id", new_column_name="job_name")
    op.rename_table("job", "runs")


def downgrade():
    op.rename_table("runs", "job")
    op.alter_column("job", "job_name", new_column_name="job_id")



"""Helper class for dbt package installation."""

from __future__ import annotations

import os
from pathlib import Path

from meltano.core import yaml
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService


class TransformAddService:
    """Helper class for adding a dbt package to the project."""

    def __init__(self, project: Project) -> None:
        """Create a new TransformAddService.

        Args:
            project: The project to add the dbt package to.
        """
        self.project = project

        self.plugins_service = ProjectPluginsService(project)

        dbt_plugin = self.plugins_service.get_transformer()

        settings_service = PluginSettingsService(
            project, dbt_plugin, plugins_service=self.plugins_service
        )
        dbt_project_dir = settings_service.get("project_dir")
        dbt_project_path = Path(dbt_project_dir)

        self.packages_file = dbt_project_path.joinpath("packages.yml")
        self.dbt_project_file = dbt_project_path.joinpath("dbt_project.yml")

    def add_to_packages(self, plugin: ProjectPlugin) -> None:
        """Add the plugin's package to the project's `packages.yml` file.

        Args:
            plugin: The plugin to add to the project.

        Raises:
            ValueError: If the plugin is missing the git repo URL.
        """
        if not os.path.exists(self.packages_file):
            self.packages_file.touch()

        package_yaml = yaml.load(self.packages_file) or {"packages": []}

        git_repo = plugin.pip_url
        if not git_repo:
            raise ValueError(f"Missing pip_url for transform plugin '{plugin.name}'")

        revision: str | None = None
        if len(git_repo.split("@")) == 2:
            git_repo, revision = git_repo.split("@")
        for package in package_yaml["packages"]:
            same_ref = (
                package.get("git", "") == git_repo
                and package.get("revision", None) == revision
            )
            if same_ref:
                return

        package_ref = {"git": git_repo}
        if revision:
            package_ref["revision"] = revision
        package_yaml["packages"].append(package_ref)

        with open(self.packages_file, "w") as f:
            yaml.dump(package_yaml, f)

    def update_dbt_project(self, plugin: ProjectPlugin) -> None:
        """Set transform package variables in `dbt_project.yml`.

        If not already present, the package name will also be added under dbt 'models'.

        Args:
            plugin: The plugin to add to the project.
        """
        settings_service = PluginSettingsService(
            self.project, plugin, plugins_service=self.plugins_service
        )

        package_name = settings_service.get("_package_name")
        package_vars = settings_service.get("_vars")

        dbt_project_yaml = yaml.load(self.dbt_project_file)

        model_def = {}

        if package_vars:
            # Add variables scoped to the plugin's package name
            config_version = dbt_project_yaml.get("config-version", 1)
            if config_version == 1:
                model_def["vars"] = package_vars
            else:
                project_vars = dbt_project_yaml.get("vars", {})
                project_vars[package_name] = package_vars
                dbt_project_yaml["vars"] = project_vars

        # Add the package's definition to the list of models:
        dbt_project_yaml["models"][package_name] = model_def

        with open(self.dbt_project_file, "w") as f:
            yaml.dump(dbt_project_yaml, f)



"""Module for working with meltano.yml files."""
from __future__ import annotations

import copy
from typing import Iterable

from meltano.core.behavior.canonical import Canonical
from meltano.core.environment import Environment
from meltano.core.plugin import PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.schedule import Schedule
from meltano.core.task_sets import TaskSets

VERSION = 1


class MeltanoFile(Canonical):
    """Data and loading methods for meltano.yml files."""

    def __init__(
        self,
        version: int = VERSION,
        plugins: dict[str, dict] = None,
        schedules: list[dict] = None,
        environments: list[dict] = None,
        jobs: list[dict] = None,
        env: dict[str, str] = None,
        **extras,
    ):
        """Construct a new MeltanoFile object from meltano.yml file.

        Args:
            version: The meltano.yml version, currently always 1.
            plugins: Plugin configuration for this project.
            schedules: Schedule configuration for this project.
            environments: Environment configuration for this project.
            jobs: Job configuration for this project.
            env: Environment variables for this project.
            extras: Additional configuration for this project.
        """
        super().__init__(
            # Attributes will be listed in meltano.yml in this order:
            version=version,
            extras=extras,
            plugins=self.load_plugins(plugins or {}),
            schedules=self.load_schedules(schedules or []),
            environments=self.load_environments(environments or []),
            jobs=self.load_job_tasks(jobs or []),
            env=env or {},
        )

    def load_plugins(self, plugins: dict[str, dict]) -> Canonical:
        """Parse the `meltano.yml` file and return it as `ProjectPlugin` instances.

        Args:
            plugins: Dictionary of plugin configurations.

        Returns:
            New ProjectPlugin instances.
        """
        plugin_type_plugins = Canonical()

        for ptype in PluginType:
            plugin_type_plugins[ptype] = []

        # this will parse the meltano.yml file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        for plugin_type, raw_plugins in plugins.items():
            if plugin_type == PluginType.MAPPERS:
                for mapper in raw_plugins:
                    plugin_type_plugins[PluginType.MAPPERS].append(
                        ProjectPlugin(PluginType.MAPPERS, **mapper)
                    )
                    plugin_type_plugins[PluginType.MAPPERS].extend(
                        self.get_plugins_for_mappings(mapper)
                    )
            else:
                for raw_plugin in raw_plugins:
                    plugin = ProjectPlugin(PluginType(plugin_type), **raw_plugin)
                    plugin_type_plugins[plugin.type].append(plugin)

        return plugin_type_plugins

    def load_schedules(self, schedules: list[dict]) -> list[Schedule]:
        """Parse the meltano.yml file and return it as Schedule instances.

        Args:
            schedules: List of schedule configurations.

        Returns:
            List of new Schedule instances.
        """
        return list(map(Schedule.parse, schedules))

    @staticmethod
    def load_environments(environments: Iterable[dict]) -> list[Environment]:
        """Parse `Environment` objects from python objects.

        Args:
            environments: Sequence of environment dictionaries.

        Returns:
            A list of `Environment` objects.
        """
        return [Environment.parse(obj) for obj in environments]

    @staticmethod
    def load_job_tasks(jobs: Iterable[dict]) -> list[TaskSets]:
        """Parse `TaskSets` objects from python objects.

        Args:
            jobs: Sequence of job dictionaries.

        Returns:
            A list of `Job` objects.
        """
        return [TaskSets.parse(obj) for obj in jobs]

    @staticmethod
    def get_plugins_for_mappings(mapper_config: dict) -> list[ProjectPlugin]:
        """Mapper plugins are a special case. They are not a single plugin, but actually a list of plugins generated from the mapping config defined within the mapper config.

        Args:
            mapper_config: The dict representation of a mapper config found in in meltano.yml.

        Returns:
            A list of `ProjectPlugin` instances.
        """
        mapping_plugins: list[ProjectPlugin] = []
        for mapping in mapper_config.get("mappings", []):
            raw_mapping_plugin = copy.deepcopy(mapper_config)
            raw_mapping_plugin["mapping"] = True
            raw_mapping_plugin["mapping_name"] = mapping.get("name")
            raw_mapping_plugin["config"] = mapping.get("config")
            mapping_plugins.append(
                ProjectPlugin(PluginType.MAPPERS, **raw_mapping_plugin)
            )
        return mapping_plugins



from __future__ import annotations

from os import SEEK_END
from os.path import join

from meltano.core.project import Project

from .utils import enforce_secure_filename

MAX_FILE_SIZE = 4194304  # 4MB max
ALLOWED_EXTENSIONS = {"json"}


class InvalidFileTypeError(Exception):
    """Occurs when a file does not meet our file type criteria."""

    def __init__(self, file):
        self.file = file
        self.extensions = ", ".join(item.upper() for item in ALLOWED_EXTENSIONS)


class InvalidFileSizeError(Exception):
    """Occurs when a file does not conform to MAX_FILE_SIZE."""

    def __init__(self, file):
        self.file = file
        self.max_file_size = MAX_FILE_SIZE


class UploadHelper:
    def upload_file(self, directory, file):
        if not self.is_valid_type(file):
            raise InvalidFileTypeError(file)

        if not self.is_valid_size(file):
            raise InvalidFileSizeError(file)

        filename = enforce_secure_filename(file.filename)
        full_path = join(directory, filename)
        file.save(full_path)

        project = Project.find()
        return full_path.replace(f"{str(project.root)}/", "")

    def is_valid_type(self, file):
        return file.filename.rsplit(".", 1)[1].lower() in ALLOWED_EXTENSIONS

    def is_valid_size(self, file):
        # File size calculated due to multipart/form data (initial header content-length may differ when complete file data has been uploaded)
        file.seek(0, SEEK_END)
        file_size = file.tell()
        file.seek(0)  # Return to 0 so subsequent write occurs from beginning of file
        return 0 < file_size <= MAX_FILE_SIZE



"""Plugin glue code for Airflow."""
from __future__ import annotations

import configparser
import logging
import os
import subprocess

from packaging.version import Version

from meltano.core.behavior.hookable import hook
from meltano.core.error import AsyncSubprocessError
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.utils import nest

from . import BasePlugin, PluginType


class AirflowInvoker(PluginInvoker):
    """Invoker that prepares env for Airflow."""

    def env(self):
        """Environment variables for Airflow.

        Returns:
            Dictionary of environment variables.
        """
        env = super().env()

        env["AIRFLOW_HOME"] = str(self.plugin_config_service.run_dir)
        env["AIRFLOW_CONFIG"] = str(self.files["config"])

        return env


class Airflow(BasePlugin):
    """Plugin glue code for Airflow."""

    __plugin_type__ = PluginType.ORCHESTRATORS

    invoker_class = AirflowInvoker

    @property
    def config_files(self):
        """Return the configuration files required by the plugin.

        Returns:
            Dictionary of config file identifiers and filenames
        """
        return {"config": "airflow.cfg"}

    def process_config(self, flat_config):
        """Unflatten the config.

        Args:
            flat_config: the flat config

        Returns:
            unflattened config
        """
        config = {}
        for key, value in flat_config.items():
            nest(config, key, str(value))
        return config

    @staticmethod
    def update_config_file(invoker: AirflowInvoker) -> None:
        """Update airflow.cfg with plugin configuration.

        Args:
            invoker: the active PluginInvoker
        """
        airflow_cfg_path = invoker.files["config"]
        logging.debug(f"Generated default '{str(airflow_cfg_path)}'")

        # open the configuration and update it
        # now we let's update the config to use our stubs
        airflow_cfg = configparser.ConfigParser()

        with airflow_cfg_path.open() as airflow_cfg_file_to_read:
            airflow_cfg.read_file(airflow_cfg_file_to_read)
            logging.debug(f"Loaded '{str(airflow_cfg_path)}'")

        config = invoker.plugin_config_processed
        for section, section_config in config.items():
            airflow_cfg[section].update(section_config)
            logging.debug(f"\tUpdated section [{section}] with {section_config}")

        with airflow_cfg_path.open("w") as airflow_cfg_file_to_write:
            airflow_cfg.write(airflow_cfg_file_to_write)
            logging.debug(f"Saved '{str(airflow_cfg_path)}'")

    @hook("before_install")
    async def setup_env(self, *args, **kwargs):
        """Configure the env to make airflow installable without GPL deps.

        Args:
            args: Arbitrary args
            kwargs: Arbitrary kwargs
        """
        os.environ["SLUGIFY_USES_TEXT_UNIDECODE"] = "yes"

    @hook("before_configure")  # noqa: WPS217
    async def before_configure(self, invoker: AirflowInvoker, session):
        """Generate config file and keep metadata database up-to-date.

        Args:
            invoker: the active PluginInvoker
            session: metadata database session

        Raises:
            AsyncSubprocessError: if command failed to run
        """
        # generate the default `airflow.cfg`
        handle = await invoker.invoke_async(
            "--help",
            require_preparation=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )
        exit_code = await handle.wait()

        if exit_code:
            raise AsyncSubprocessError(
                "Command `airflow --help` failed", process=handle
            )

        # Read and update airflow.cfg
        self.update_config_file(invoker)

        # we've changed the configuration here, so we need to call
        # prepare again on the invoker so it re-reads the configuration
        # for the Airflow plugin
        await invoker.prepare(session)

        # make sure we use correct db init
        handle = await invoker.invoke_async(
            "version",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        stdout, stderr = await handle.communicate()

        if handle.returncode:
            raise AsyncSubprocessError(
                "Command `airflow version` failed", process=handle
            )

        version = stdout.decode()
        init_db_cmd = (
            ["initdb"] if Version(version) < Version("2.0.0") else ["db", "init"]
        )

        handle = await invoker.invoke_async(
            *init_db_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        exit_code = await handle.wait()

        if exit_code:
            raise AsyncSubprocessError(
                "Airflow metadata database could not be initialized: `airflow initdb` failed",
                handle,
            )

        logging.debug("Completed `airflow initdb`")

    @hook("before_cleanup")
    async def before_cleanup(self, invoker: PluginInvoker):
        """Delete the config file.

        Args:
            invoker: the active PluginInvoker
        """
        config_file = invoker.files["config"]
        try:
            config_file.unlink()
            logging.debug(f"Deleted configuration at {config_file}")
        except FileNotFoundError:
            pass



"""Basic `MailService` that sends notifications via the sendgrid API."""

from __future__ import annotations

from flask_mail import Mail, Message
from smtpapi import SMTPAPIHeader

from meltano.api.models.subscription import Subscription, SubscriptionEventType
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService

mail = Mail()


class MailService:
    """Basic MailService that sends notifications via the sendgrid SMTPAPI."""

    def __init__(self, project: Project):
        """Initialize the MailService.

        Args:
            project: The project to use the MailService when referencing the project settings or project id.
        """
        self.project = project
        self._settings = ProjectSettingsService(self.project)

        self.project_id = self._settings.get("project_id")
        self.sendgrid_unsubscribe_group_id = self._settings.get(
            "mail.sendgrid_unsubscribe_group_id"
        )

    def get_unsubscribe_group(self, subscription: Subscription) -> int | None:
        """Get the unsubscribe group for the given subscription.

        Currently, the unsubscribe group is based on the event type, and only the SubscriptionEventType.PIPELINE_MANUAL_RUN
        even type is actually supported.

        Args:
            subscription: The subscription to get the unsubscribe group for.

        Returns:
            The unsubscribe group id or None if the event type is not supported.
        """
        event_type = SubscriptionEventType(subscription.event_type)
        if event_type == SubscriptionEventType.PIPELINE_MANUAL_RUN:
            return self.sendgrid_unsubscribe_group_id
        return None

    def create_message(self, subscription: Subscription, **kwargs) -> Message:
        """Create a message for the given subscription.

        Args:
            subscription: The subscription to create the message for.
            kwargs: Additional arguments to pass to the message constructor.

        Returns:
            The created message.
        """
        headers = SMTPAPIHeader()
        headers.set_unique_args(
            {
                "project_id": str(self.project_id),
                "subscription_id": str(subscription.id),
            }
        )

        unsubscribe_group = self.get_unsubscribe_group(subscription)
        if unsubscribe_group:
            headers.set_asm_group_id(int(unsubscribe_group))

        return Message(
            recipients=[subscription.recipient],
            extra_headers={"X-SMTPAPI": headers.json_string()},
            **kwargs,
        )



"""StateStoreManager for S3 cloud storage backend."""
from __future__ import annotations

import re

from meltano.core.state_store.filesystem import (
    BaseFilesystemStateStoreManager,
    InvalidStateBackendConfigurationException,
)


class S3StateStoreManager(BaseFilesystemStateStoreManager):
    """State backend for S3."""

    label: str = "AWS S3"

    def __init__(
        self,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        bucket: str | None = None,
        prefix: str | None = None,
        endpoint_url: str | None = None,
        **kwargs,
    ):
        """Initialize the BaseFilesystemStateStoreManager.

        Args:
            aws_access_key_id: AWS access key ID used to authenticate
            aws_secret_access_key: AWS secret access key used to authenticate
            bucket: the bucket to store state in
            prefix: the prefix to store state at
            endpoint_url: the endpoint url for S3
            kwargs: additional keyword args to pass to parent
        """
        super().__init__(**kwargs)
        self.aws_access_key_id = aws_access_key_id or self.parsed.username
        self.aws_secret_access_key = aws_secret_access_key or self.parsed.password
        self.bucket = bucket or self.parsed.hostname
        self.prefix = prefix or self.parsed.path
        self.endpoint_url = endpoint_url
        self._client = None

    @staticmethod
    def is_file_not_found_error(err: Exception) -> bool:
        """Check if err is equivalent to file not being found.

        Args:
            err: the err to check

        Returns:
            True if error represents file not being found, else False
        """
        return (
            isinstance(err, OSError)
            and "NoSuchKey" in err.args[0]
            or (isinstance(err, KeyError) and "ActualObjectSize" in err.args[0])
        )

    @property
    def client(self):
        """Get an authenticated boto3.Client.

        Returns:
            A boto3.Client.

        Raises:
            InvalidStateBackendConfigurationException: when configured AWS settings are invalid.
        """
        if not self._client:
            if self.aws_secret_access_key and self.aws_access_key_id:
                from boto3 import Session

                session = Session(
                    aws_access_key_id=self.aws_access_key_id,
                    aws_secret_access_key=self.aws_secret_access_key,
                )
                self._client = session.client("s3", endpoint_url=self.endpoint_url)
            elif self.aws_secret_access_key and not self.aws_access_key_id:
                raise InvalidStateBackendConfigurationException(
                    "AWS secret access key configured, but not AWS access key ID."
                )
            elif self.aws_access_key_id and not self.aws_secret_access_key:
                raise InvalidStateBackendConfigurationException(
                    "AWS access key ID configured, but no AWS secret access key."
                )
        return self._client

    @property
    def state_dir(self) -> str:
        """Get the prefix that state should be stored at.

        Returns:
            The relevant prefix
        """
        return self.prefix.lstrip(self.delimiter).rstrip(self.delimiter)

    def get_state_ids(self, pattern: str | None = None):
        """Get list of state_ids stored in the backend.

        Args:
            pattern: glob-style pattern to filter state_ids by

        Returns:
            List of state_ids
        """
        if pattern:
            pattern_re = re.compile(pattern.replace("*", ".*"))
        state_ids = set()
        for state_obj in self.client.list_objects_v2(
            Bucket=self.bucket, Prefix=self.prefix
        ).get("Contents", []):
            (state_id, filename) = state_obj["Key"].split("/")[-2:]
            if filename == "state.json":
                if not pattern:
                    state_ids.add(
                        state_id.replace(self.prefix, "").replace("/state.json", "")
                    )
                elif pattern_re.match(state_id):
                    state_ids.add(state_id)
        return list(state_ids)

    def delete(self, file_path: str):
        """Delete the file/blob at the given path.

        Args:
            file_path: the path to delete.
        """
        self.client.delete_objects(
            Bucket=self.bucket,
            Delete={"Objects": [{"Key": file_path}]},
        )



"""New Project Initialization CLI."""

from __future__ import annotations

import logging
from pathlib import Path

import click

from meltano.cli import cli
from meltano.cli.params import database_uri_option
from meltano.cli.utils import CliError, InstrumentedCmd
from meltano.core.project_init_service import ProjectInitService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.tracking import Tracker
from meltano.core.tracking.contexts import CliContext, CliEvent

EXTRACTORS = "extractors"
LOADERS = "loaders"
ALL = "all"

logger = logging.getLogger(__name__)
path_type = click.Path(file_okay=False, path_type=Path)


@cli.command(cls=InstrumentedCmd, short_help="Create a new Meltano project.")
@click.pass_context
@click.argument("project_directory", required=False, type=path_type)
@click.option(
    "--no_usage_stats", help="Do not send anonymous usage stats.", is_flag=True
)
@database_uri_option
def init(ctx, project_directory: Path, no_usage_stats):
    """
    Create a new Meltano project.

    Read more at https://docs.meltano.com/reference/command-line-interface#init

    """
    if not project_directory:
        click.echo("We need a project name to get started!")
        project_directory = click.prompt(
            "Enter a name now to create a Meltano project", type=path_type
        )

    if ctx.obj["project"]:
        root = ctx.obj["project"].root
        logging.warning(f"Found meltano project at: {root}")
        raise CliError("`meltano init` cannot run inside a Meltano project.")

    if no_usage_stats:
        ProjectSettingsService.config_override["send_anonymous_usage_stats"] = False

    init_service = ProjectInitService(project_directory)

    project = init_service.init()
    init_service.echo_instructions(project)

    # since the project didn't exist, tracking was not initialized in cli.py
    # but now that we've created a project we can set up telemetry and fire events.
    tracker = Tracker(project)
    tracker.add_contexts(CliContext.from_click_context(ctx))
    tracker.track_command_event(CliEvent.started)
    ctx.obj["tracker"] = tracker



"""CLI command `meltano invoke`."""

from __future__ import annotations

import asyncio
import logging
import sys

import click
from sqlalchemy.orm import sessionmaker

from meltano.cli import cli
from meltano.cli.params import pass_project
from meltano.cli.utils import (
    CliEnvironmentBehavior,
    CliError,
    PartialInstrumentedCmd,
    propagate_stop_signals,
)
from meltano.core.db import project_engine
from meltano.core.error import AsyncSubprocessError
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin_invoker import (
    PluginInvoker,
    UnknownCommandError,
    invoker_factory,
)
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.tracking import Tracker
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext

logger = logging.getLogger(__name__)


@cli.command(
    cls=PartialInstrumentedCmd,
    context_settings={"ignore_unknown_options": True, "allow_interspersed_args": False},
    short_help="Invoke a plugin.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_use_default,
)
@click.option(
    "--print-var",
    help="Print to stdout the values for the provided environment variables, as passed to the plugininvoker context. Useful for debugging.",
    multiple=True,
)
@click.option(
    "--plugin-type", type=click.Choice(PluginType.cli_arguments()), default=None
)
@click.option(
    "--dump",
    type=click.Choice(["catalog", "config"]),
    help="Dump content of specified file to disk.",
)
@click.option(
    "--list-commands",
    is_flag=True,
    help="List the commands supported by the plugin.",
)
@click.argument("plugin_name", metavar="PLUGIN_NAME[:COMMAND_NAME]")
@click.argument("plugin_args", nargs=-1, type=click.UNPROCESSED)
@click.option(
    "--containers",
    is_flag=True,
    help="Execute plugins using containers where possible.",
)
@click.pass_context
@pass_project(migrate=True)
def invoke(
    project: Project,
    ctx: click.Context,
    plugin_type: str,
    dump: str,
    list_commands: bool,
    plugin_name: str,
    plugin_args: tuple[str, ...],
    containers: bool = False,
    print_var: str | None = None,
):
    """
    Invoke a plugin's executable with specified arguments.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#invoke
    """
    tracker: Tracker = ctx.obj["tracker"]

    try:
        plugin_name, command_name = plugin_name.split(":")
    except ValueError:
        command_name = None

    plugin_type = PluginType.from_cli_argument(plugin_type) if plugin_type else None

    _, Session = project_engine(project)  # noqa: N806
    session = Session()
    plugins_service = ProjectPluginsService(project)

    try:
        plugin = plugins_service.find_plugin(
            plugin_name, plugin_type=plugin_type, invokable=True
        )
        tracker.add_contexts(PluginsTrackingContext([(plugin, command_name)]))
        tracker.track_command_event(CliEvent.inflight)
    except PluginNotFoundError:
        tracker.track_command_event(CliEvent.aborted)
        raise

    if list_commands:
        do_list_commands(plugin)
        tracker.track_command_event(CliEvent.completed)
        return

    invoker = invoker_factory(project, plugin, plugins_service=plugins_service)
    try:
        exit_code = asyncio.run(
            _invoke(
                invoker,
                project,
                plugin_name,
                plugin_args,
                session,
                dump,
                command_name,
                containers,
                print_var=print_var,
            )
        )
    except Exception as invoke_err:
        tracker.track_command_event(CliEvent.failed)
        raise invoke_err

    if exit_code == 0:
        tracker.track_command_event(CliEvent.completed)
    else:
        tracker.track_command_event(CliEvent.failed)
    sys.exit(exit_code)


async def _invoke(
    invoker: PluginInvoker,
    project: Project,
    plugin_name: str,
    plugin_args: str,
    session: sessionmaker,
    dump: str,
    command_name: str,
    containers: bool,
    print_var: list | None = None,
):
    if command_name is not None:
        command = invoker.find_command(command_name)

    try:
        async with invoker.prepared(session):
            if print_var:
                env = invoker.env()
                for key in print_var:
                    val = env.get(key)
                    click.echo(f"{key}={val}")
            if dump:
                await dump_file(invoker, dump)
                exit_code = 0
            elif (  # noqa: WPS337
                containers
                and command_name is not None
                and command.container_spec is not None
            ):
                return await invoker.invoke_docker(
                    command_name,
                    *plugin_args,
                )
            else:
                handle = await invoker.invoke_async(*plugin_args, command=command_name)
                with propagate_stop_signals(handle):
                    exit_code = await handle.wait()

    except UnknownCommandError as err:
        raise click.BadArgumentUsage(err) from err
    except AsyncSubprocessError as err:
        logger.error(await err.stderr)
        raise
    finally:
        session.close()

    return exit_code


def do_list_commands(plugin):
    """List the commands supported by plugin."""
    if not plugin.supported_commands:
        click.secho(
            f"Plugin '{plugin.name}' does not define any commands.", fg="yellow"
        )
        return

    descriptions = {
        f"{plugin.name}:{cmd}": props.description
        for cmd, props in plugin.all_commands.items()
    }
    column_len = max(len(name) for name in descriptions.keys()) + 2
    for name, desc in descriptions.items():
        click.secho(name.ljust(column_len, " "), fg="blue", nl=False)
        click.echo(desc)


async def dump_file(invoker: PluginInvoker, file_id: str):
    """Dump file."""
    try:
        content = await invoker.dump(file_id)
    except FileNotFoundError as err:
        raise CliError(f"Could not find {file_id}") from err
    except Exception as err:
        raise CliError(f"Could not dump {file_id}: {err}") from err
    print(content)  # noqa: WPS421



"""Various utilities for configuring logging in a meltano project."""

from __future__ import annotations

import asyncio
import logging
import os
import sys
from contextlib import suppress
from logging import config as logging_config

import structlog
import yaml

from meltano.core.logging.formatters import (
    LEVELED_TIMESTAMPED_PRE_CHAIN,
    TIMESTAMPER,
    rich_exception_formatter_factory,
)
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import get_no_color_flag

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


LEVELS = {  # noqa: WPS407
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
DEFAULT_LEVEL = "info"
FORMAT = "[%(asctime)s] [%(process)d|%(threadName)10s|%(name)s] [%(levelname)s] %(message)s"  # noqa: WPS323


def parse_log_level(log_level: dict[str, int]) -> int:
    """Parse a level descriptor into an logging level.

    Args:
        log_level: level descriptor.

    Returns:
        int: actual logging level.
    """
    return LEVELS.get(log_level, LEVELS[DEFAULT_LEVEL])


def read_config(config_file: str | None = None) -> dict:
    """Read a logging config yaml from disk.

    Args:
        config_file: path to the config file to read.

    Returns:
        dict: parsed yaml config
    """
    if config_file and os.path.exists(config_file):
        with open(config_file) as cf:
            return yaml.safe_load(cf.read())
    else:
        return None


def default_config(log_level: str) -> dict:
    """Generate a default logging config.

    Args:
        log_level: set log levels to provided level.

    Returns:
         A logging config suitable for use with `logging.config.dictConfig`.
    """
    no_color = get_no_color_flag()

    if no_color:
        formatter = rich_exception_formatter_factory(no_color=True)
    else:
        formatter = rich_exception_formatter_factory(color_system="truecolor")

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "colored": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.dev.ConsoleRenderer(
                    colors=not no_color,
                    exception_formatter=formatter,
                ),
                "foreign_pre_chain": LEVELED_TIMESTAMPED_PRE_CHAIN,
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": log_level.upper(),
                "formatter": "colored",
                "stream": "ext://sys.stderr",
            },
        },
        "loggers": {
            "": {
                "handlers": ["console"],
                "level": log_level.upper(),
                "propagate": True,
            },
            "snowplow_tracker.emitters": {
                "handlers": ["console"],
                "level": logging.ERROR,
            },
            "urllib3": {
                "handlers": ["console"],
                "level": logging.INFO,
                "propagate": False,
            },
            "asyncio": {
                "handlers": ["console"],
                "level": logging.INFO,
                "propagate": False,
            },
        },
    }


def setup_logging(  # noqa: WPS210
    project: Project | None = None,
    log_level: str = DEFAULT_LEVEL,
    log_config: dict | None = None,
) -> None:
    """Configure logging for a meltano project.

    Args:
        project: the meltano project
        log_level: set log levels to provided level.
        log_config: a logging config suitable for use with `logging.config.dictConfig`.
    """
    # Mimick Python 3.8's `force=True` kwarg to override any
    # existing logger handlers
    # See https://github.com/python/cpython/commit/cf67d6a934b51b1f97e72945b596477b271f70b8
    root = logging.getLogger()
    for handler in root.handlers[:]:
        root.removeHandler(handler)
        handler.close()

    log_level = log_level.upper()

    if project:
        settings_service = ProjectSettingsService(project)
        log_config = log_config or settings_service.get("cli.log_config")
        log_level = settings_service.get("cli.log_level")

    config = read_config(log_config) or default_config(log_level)
    logging_config.dictConfig(config)
    structlog.configure(
        processors=[
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            TIMESTAMPER,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def change_console_log_level(log_level: int = logging.DEBUG) -> None:
    """Change the log level for the current root logger, but only on the 'console' handler.

    Most useful when you want change the log level on the fly for console output, but want to respect other aspects
    of any potential logging.yaml sourced configs. Note that if a logging.yaml config without a 'console' handler
    is used, this will not override the log level.

    Args:
        log_level: set log levels to provided level.
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)
    for handler in root_logger.handlers:
        if handler.name == "console":
            handler.setLevel(log_level)


class SubprocessOutputWriter(Protocol):
    """SubprocessOutputWriter is a basic interface definition suitable for use with capture_subprocess_output."""

    def writelines(self, lines: str):
        """Any type with a writelines method accepting a string could be used as an output writer.

        Args:
            lines: string to write
        """


async def _write_line_writer(writer, line):
    # StreamWriters like a subprocess's stdin need special consideration
    if isinstance(writer, asyncio.StreamWriter):
        try:  # noqa: WPS229
            writer.write(line)
            await writer.drain()
        except (BrokenPipeError, ConnectionResetError):
            with suppress(AttributeError):  # `wait_closed` is Python 3.7+
                await writer.wait_closed()

            return False
    else:
        writer.writeline(line.decode())

    return True


async def capture_subprocess_output(
    reader: asyncio.StreamReader | None, *line_writers: SubprocessOutputWriter
) -> None:
    """Capture in real time the output stream of a suprocess that is run async.

    The stream has been set to asyncio.subprocess.PIPE and is provided using
    reader to this function.

    As new lines are captured for reader, they are written to output_stream.
    This async function should be run with await asyncio.wait() while waiting
    for the subprocess to end.

    Args:
        reader: asyncio.StreamReader object that is the output stream of the subprocess.
        line_writers: any object thats a StreamWriter or has a writelines method accepting a string.
    """
    while not reader.at_eof():
        line = await reader.readline()
        if not line:
            continue

        for writer in line_writers:
            if not await _write_line_writer(writer, line):
                # If the destination stream is closed, we can stop capturing output.
                return



"""Meltano UI worker thread definition."""

from __future__ import annotations

import logging
import threading
import time
import traceback

import click
import requests

from meltano.core.project_settings_service import ProjectSettingsService

logger = logging.getLogger(__name__)

SUCCESS_STATUS_CODE = 200


class UIAvailableWorker(threading.Thread):
    """A thread subclass for Meltano UI workers."""

    def __init__(self, project):
        """Initialize the `UIAvailableWorker` thread.

        Args:
            project: The Meltano project.
        """
        super().__init__()
        self.project = project
        self.settings_service = ProjectSettingsService(self.project)
        self._terminate = False

    def run(self) -> None:
        """Run the thread, and report when the Meltano UI becomes available."""
        url = f"http://localhost:{self.settings_service.get('ui.bind_port')}"
        headers = {"Host": self.settings_service.get("ui.server_name")}

        while not self._terminate:
            try:
                response = requests.get(url, headers=headers)
                if response.status_code == SUCCESS_STATUS_CODE:
                    click.secho(f"Meltano UI is now available at {url}", fg="green")
                    self._terminate = True
            except Exception:
                logger.debug(
                    f"Exception encountered while trying to run Meltano UI:\n{traceback.format_exc()}"
                )

            time.sleep(2)

    def stop(self):
        """Stop the thread."""
        self._terminate = True



"""Flask app for Meltano UI."""
from __future__ import annotations

import importlib
import logging
from urllib.parse import urlsplit

from flask import Flask, g, jsonify, request  # noqa: WPS347
from flask_cors import CORS
from werkzeug.middleware.dispatcher import DispatcherMiddleware

from meltano import __version__ as meltano_version
from meltano.api import config as api_config
from meltano.api.headers import VERSION_HEADER
from meltano.api.security.auth import HTTP_READONLY_CODE
from meltano.core.db import project_engine
from meltano.core.logging.utils import FORMAT, setup_logging
from meltano.core.project import Project, ProjectReadonly
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.oauth.app import create_app as create_oauth_service

STATUS_SERVER_ERROR = 500

setup_logging()

# the file logger we will set up below is for the `meltano.api` module
logger = logging.getLogger("meltano.api")


def create_app(config: dict = {}) -> Flask:  # noqa: WPS210,WPS213,B006
    """Create flask app for the current project.

    Args:
        config: app configuration

    Returns:
        Flask app
    """
    project = Project.find()
    setup_logging(project)

    settings_service = ProjectSettingsService(project)

    project_engine(project, default=True)

    app = Flask(
        __name__, instance_path=str(project.root), instance_relative_config=True
    )

    # make sure we have the latest environment loaded
    importlib.reload(api_config)

    app.config.from_object("meltano.api.config")
    app.config.from_mapping(**api_config.ProjectSettings(project).as_dict())
    app.config.from_mapping(**config)

    # File logging
    file_handler = logging.handlers.RotatingFileHandler(
        str(project.run_dir("meltano-ui.log")), backupCount=3
    )
    formatter = logging.Formatter(fmt=FORMAT)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    # 1) Extensions
    from .executor import setup_executor
    from .json import setup_json
    from .mail import mail
    from .models import db
    from .security import setup_security
    from .security.oauth import setup_oauth

    db.init_app(app)
    mail.init_app(app)
    setup_executor(app, project)
    setup_security(app, project)
    setup_oauth(app)
    setup_json(app)

    # we need to setup CORS for development
    if app.env == "development":
        CORS(app, origins="http://localhost:8080", supports_credentials=True)

    # 2) Register the URL Converters
    from .url_converters import PluginRefConverter

    app.url_map.converters["plugin_ref"] = PluginRefConverter

    # 3) Register the controllers

    from .controllers.orchestrations import orchestrations_bp
    from .controllers.plugins import plugins_bp
    from .controllers.root import api_root, root
    from .controllers.settings import settings_bp

    app.register_blueprint(settings_bp)
    app.register_blueprint(orchestrations_bp)
    app.register_blueprint(plugins_bp)
    app.register_blueprint(root)
    app.register_blueprint(api_root)

    if app.config["PROFILE"]:
        from .profiler import init

        init(app)

    # Notifications
    if settings_service.get("ui.notification"):
        from .events import notifications

        notifications.init_app(app)
        logger.debug("Notifications are enabled.")
    else:
        logger.debug("Notifications are disabled.")

    @app.before_request
    def setup_js_context():
        # setup the appUrl
        g.jsContext = {
            "appUrl": urlsplit(request.host_url).geturl()[:-1],
            "version": meltano_version,
        }

        setting_map = {
            "isSendAnonymousUsageStats": "send_anonymous_usage_stats",
            "projectId": "project_id",
            "isProjectReadonlyEnabled": "project_readonly",
            "isReadonlyEnabled": "ui.readonly",
            "isAnonymousReadonlyEnabled": "ui.anonymous_readonly",
            "isNotificationEnabled": "ui.notification",
            "isAnalysisEnabled": "ui.analysis",
            "logoUrl": "ui.logo_url",
            "oauthServiceUrl": "oauth_service.url",
        }

        for context_key, setting_name in setting_map.items():
            g.jsContext[context_key] = settings_service.get(setting_name)

        providers = settings_service.get("oauth_service.providers")
        g.jsContext["oauthServiceProviders"] = [
            provider for provider in providers.split(",") if provider
        ]

    @app.after_request
    def after_request(res):
        res.headers[VERSION_HEADER] = meltano_version
        return res

    @app.errorhandler(STATUS_SERVER_ERROR)
    def internal_error(exception):
        logger.info(f"Error: {exception}")
        return jsonify({"error": True, "code": str(exception)}), STATUS_SERVER_ERROR

    @app.errorhandler(ProjectReadonly)
    def _handle(ex):
        return (jsonify({"error": True, "code": str(ex)}), HTTP_READONLY_CODE)

    # create the dispatcher to host the `OAuthService`
    app.wsgi_app = DispatcherMiddleware(
        app.wsgi_app, {"/-/oauth": create_oauth_service()}
    )

    return app



"""User account management CLI."""

from __future__ import annotations

import click

from meltano.api.app import create_app

from . import cli
from .params import pass_project
from .utils import InstrumentedCmd, InstrumentedGroup


@cli.group(
    cls=InstrumentedGroup,
    invoke_without_command=True,
    short_help="Manage Meltano user accounts.",
)
@pass_project(migrate=True)
@click.pass_context
def user(ctx, project):
    """
    Manage Meltano user accounts.

    TIP: This command is only relevant when Meltano is run with authentication enabled.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#user
    """
    ctx.obj["project"] = project


@user.command(cls=InstrumentedCmd, short_help="Create a Meltano user account.")
@click.argument("username")
@click.argument("password")
@click.option(
    "--overwrite",
    "-f",
    is_flag=True,
    default=False,
    help="Update the user instead of creating a new one.",
)
@click.option(
    "--role",
    "-G",
    multiple=True,
    default=[],
    help="Add the user to the role. Meltano ships with two built-in roles: admin and regular.",
)
@click.pass_context
def add(ctx, username, password, role, **flags):
    """
    Create a Meltano user account.

    TIP: This command is only relevant when Meltano is run with authentication enabled.
    """
    app = create_app()

    from flask_security.utils import hash_password

    from meltano.api.security.identity import users

    try:
        with app.app_context():
            # make sure our User doesn't already exist
            if not flags["overwrite"] and users.get_user(username):
                raise Exception(
                    f"User '{username}' already exists. Use --overwrite to update it."
                )

            # make sure all roles exists
            roles = []
            for role_name in role:
                this_role = users.find_role(role_name)
                if not this_role:
                    raise Exception(f"Role '{role_name}' does not exists.")

                roles.append(this_role)

            current_user = users.get_user(username) or users.create_user(
                username=username
            )

            current_user.password = hash_password(password)
            current_user.roles = roles

            # for some reason the scoped_session doesn't trigger the commit
            users.db.session.commit()
    except Exception as err:
        click.secho(f"Could not create user '{username}': {err}", fg="red")
        raise click.Abort from err



"""Add plugins to the project."""

from __future__ import annotations

import enum

from .plugin import BasePlugin, PluginType, Variant
from .plugin.project_plugin import ProjectPlugin
from .project import Project
from .project_plugins_service import PluginAlreadyAddedException, ProjectPluginsService


class PluginAddedReason(str, enum.Enum):
    """The reason why a plugin was added to the project."""

    #: The plugin was added by the user.
    USER_REQUEST = "user_request"

    #: The plugin was added because it is related to another plugin.
    RELATED = "related"

    #: The plugin was added because it is required by another plugin.
    REQUIRED = "required"


class MissingPluginException(Exception):
    """Raised when a plugin is not found."""


class ProjectAddService:
    """Project Add Service."""

    def __init__(
        self,
        project: Project,
        plugins_service: ProjectPluginsService | None = None,
    ):
        """Create a new Project Add Service.

        Args:
            project: The project to add plugins to.
            plugins_service: The project plugins service.
        """
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)

    def add(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        lock: bool = True,
        **attrs,
    ) -> ProjectPlugin:
        """Add a new plugin to the project.

        Args:
            plugin_type: The type of the plugin to add.
            plugin_name (str): The name of the plugin to add.
            lock: Whether to generate a lockfile for the plugin.
            attrs: Additional attributes to add to the plugin.

        Returns:
            The added plugin.
        """
        plugin = ProjectPlugin(
            plugin_type, plugin_name, **attrs, default_variant=Variant.DEFAULT_NAME
        )

        with self.plugins_service.disallow_discovery_yaml():
            self.plugins_service.ensure_parent(plugin)

            # If we are inheriting from a base plugin definition,
            # repeat the variant and pip_url in meltano.yml
            parent = plugin.parent
            if isinstance(parent, BasePlugin):
                plugin.variant = parent.variant
                plugin.pip_url = parent.pip_url

            added = self.add_plugin(plugin)

            if lock and not added.is_custom():
                self.plugins_service.lock_service.save(
                    added,
                    exists_ok=plugin.inherit_from is not None,
                )

            return added

    def add_plugin(self, plugin: ProjectPlugin):
        """Add a plugin to the project.

        Args:
            plugin: The plugin to add.

        Returns:
            The added plugin.
        """
        return self.plugins_service.add_to_file(plugin)

    def add_required(
        self,
        plugin: ProjectPlugin,
        lock: bool = True,
    ):
        """Add all required plugins to the project.

        Args:
            plugin: The plugin to get requirements from.
            lock: Whether to generate a lockfile for the plugin.

        Returns:
            The added plugins.
        """
        added_plugins = []
        for plugin_type, plugins in plugin.all_requires.items():
            for plugin_req in plugins:
                try:
                    plugin = self.add(
                        plugin_type,
                        plugin_req.name,
                        variant=plugin_req.variant,
                        lock=lock,
                    )
                except PluginAlreadyAddedException:
                    continue

                added_plugins.append(plugin)

        added_plugins_with_required = []
        for added in added_plugins:
            added_plugins_with_required.extend([added, *self.add_required(added)])

        return added_plugins_with_required



from __future__ import annotations

from flask_login import user_logged_in
from flask_principal import identity_loaded

from meltano.core.project_settings_service import ProjectSettingsService

from .auth import (  # noqa: WPS450
    _identity_loaded_hook,
    _user_logged_in_hook,
    unauthorized_callback,
)


def setup_security(app, project):
    from meltano.api.security.forms import (
        MeltanoConfirmRegisterForm,
        MeltanoLoginForm,
        MeltanoRegisterFrom,
    )
    from meltano.api.security.identity import FreeUser, users

    options = {
        "login_form": MeltanoLoginForm,
        "register_form": MeltanoRegisterFrom,
        "confirm_register_form": MeltanoConfirmRegisterForm,
    }

    settings_service = ProjectSettingsService(project)
    if not settings_service.get("ui.authentication"):
        options["anonymous_user"] = FreeUser
    # Else: use Flask's built-in AnonymousUser, which is not deemed to be
    # authenticated and has no roles.

    from flask_security import Security

    security = Security()
    # normally one would setup the extension accordingly, but it
    # seems Security.init_app() overwrites all the configuration
    security.init_app(app, users, **options)
    security.unauthorized_handler(unauthorized_callback)
    user_logged_in.connect_via(app)(_user_logged_in_hook)
    identity_loaded.connect_via(app)(_identity_loaded_hook)



"""Meltano Setting Definitions."""

from __future__ import annotations

import json
from datetime import date, datetime
from enum import Enum
from typing import Any

from ruamel.yaml import Representer

from meltano.core import utils
from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical
from meltano.core.error import Error

VALUE_PROCESSORS = {
    "nest_object": utils.nest_object,
    "upcase_string": lambda vlu: vlu.upper(),
    "stringify": lambda vlu: vlu if isinstance(vlu, str) else json.dumps(vlu),
}


class EnvVar:
    """Environment Variable class."""

    def __init__(self, definition: str):
        """Instantiate new EnvVar.

        Args:
            definition: Env var definition.
        """
        key = definition
        negated = False

        if definition.startswith("!"):
            key = definition[1:]
            negated = True

        self.key = key
        self.negated = negated

    @property
    def definition(self) -> str:
        """Return env var definition.

        Returns:
            Env var definition.
        """
        prefix = "!" if self.negated else ""
        return f"{prefix}{self.key}"

    def get(self, env) -> str:
        """Get env value.

        Args:
            env: Env to get value for.

        Returns:
            Env var value for given env var.
        """
        if self.negated:
            return str(not utils.truthy(env[self.key]))
        return env[self.key]


class SettingMissingError(Error):
    """Occurs when a setting is missing."""

    def __init__(self, name: str):
        """Instantiate SettingMissingError.

        Args:
            name: Name of missing setting.
        """
        super().__init__(f"Cannot find setting {name}")


class YAMLEnum(str, Enum):
    """Serializable Enum class."""

    def __str__(self):
        """Return as string.

        Returns:
            This enum in string form.
        """
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj) -> str:
        """Represent as yaml.

        Args:
            dumper: YAML dumper.
            obj: Object to dump.

        Returns:
            Object in yaml string form.
        """
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))

    @classmethod
    def to_yaml(cls, representer: Representer, node: Any):
        """Represent as yaml.

        Args:
            representer: YAML representer.
            node: Object to dump.

        Returns:
            Object in yaml string form.
        """
        return representer.represent_scalar("tag:yaml.org,2002:str", str(node))

    @classmethod
    def from_yaml(cls, constructor, node):
        """Construct from yaml.

        Args:
            constructor: Class constructor.
            node: YAML node.

        Returns:
            Object from yaml node.
        """
        return cls(node.value)


class SettingKind(YAMLEnum):
    """Supported setting kinds."""

    STRING = "string"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    DATE_ISO8601 = "date_iso8601"
    EMAIL = "email"
    PASSWORD = "password"  # noqa: S105
    OAUTH = "oauth"
    OPTIONS = "options"
    FILE = "file"
    ARRAY = "array"
    OBJECT = "object"
    HIDDEN = "hidden"


class SettingDefinition(NameEq, Canonical):
    """Meltano SettingDefinition class."""

    def __init__(
        self,
        name: str | None = None,
        aliases: list[str] = None,
        env: str | None = None,
        env_aliases: list[str] = None,
        kind: SettingKind | None = None,
        value=None,
        label: str | None = None,
        documentation: str | None = None,
        description: str | None = None,
        tooltip: str | None = None,
        options: list | None = None,
        oauth: dict | None = None,
        placeholder: str | None = None,
        protected: bool | None = None,
        env_specific: bool | None = None,
        custom: bool = False,
        value_processor=None,
        value_post_processor=None,
        **attrs,
    ):
        """Instantiate new SettingDefinition.

        Args:
            name: Setting name.
            aliases: Setting alias names.
            env: Setting target environment variable.
            env_aliases: Deprecated. Used to delegate alternative environment variables for overriding this setting's value.
            kind: Setting kind.
            value: Setting value.
            label: Setting label.
            documentation: Setting docs url.
            description: Setting description.
            tooltip: A phrase to provide additional information on this setting.
            options: Setting options.
            oauth: Setting OAuth provider details.
            placeholder: A placeholder value for this setting.
            protected: A protected setting cannot be changed from the UI.
            env_specific: Flag for environment-specific setting.
            custom: Custom setting flag.
            value_processor: Used with `kind: object` to pre-process the keys in a particular way.
            value_post_processor: Used with `kind: object` to post-process the keys in a particular way.
            attrs: Keyword arguments to pass to parent class.
        """
        aliases = aliases or []
        env_aliases = env_aliases or []
        options = options or []
        oauth = oauth or {}

        super().__init__(
            # Attributes will be listed in meltano.yml in this order:
            name=name,
            aliases=aliases,
            env=env,
            env_aliases=env_aliases,
            kind=SettingKind(kind) if kind else None,
            value=value,
            label=label,
            documentation=documentation,
            description=description,
            tooltip=tooltip,
            options=options,
            oauth=oauth,
            placeholder=placeholder,
            protected=protected,
            env_specific=env_specific,
            value_processor=value_processor,
            value_post_processor=value_post_processor,
            _custom=custom,
            **attrs,
        )

        self._verbatim.add("value")

    def __repr__(self) -> str:
        """Return string representation.

        Returns:
            String representation of this setting.
        """
        return f"<SettingDefinition {self.name} ({self.kind})>"

    @classmethod
    def from_missing(cls, defs: list[SettingDefinition], config: dict, **kwargs):
        """Create SettingDefinition instances for missing settings.

        Args:
            defs: Know setting definitions.
            config: Config dict.
            kwargs: Keyword arguments to pass to new SettingDefinition instances.

        Returns:
            A list of created custom SettingDefinition instances.
        """
        flat_config = utils.flatten(config, "dot")

        names = {setting.name for setting in defs}

        # Create custom setting definitions for unknown keys
        return [
            SettingDefinition.from_key_value(key, value, **kwargs)
            for key, value in flat_config.items()
            if key not in names
        ]

    @classmethod
    def from_key_value(
        cls,
        key: str,
        value: Any,
        custom: bool = True,
        default: Any | bool = False,
    ):
        """Create SettingDefinition instance from key-value pair.

        Args:
            key: Key.
            value: Value.
            custom: Custom setting flag.
            default: Default setting value.

        Returns:
            A SettingDefinition instance.
        """
        kind = None
        if isinstance(value, bool):
            kind = SettingKind.BOOLEAN
        elif isinstance(value, int):
            kind = SettingKind.INTEGER
        elif isinstance(value, dict):
            kind = SettingKind.OBJECT
        elif isinstance(value, list):
            kind = SettingKind.ARRAY

        attrs = {
            "name": key,
            "kind": kind,
            "custom": custom,
            "value": value if default else None,
        }

        return cls(**attrs)

    @property
    def is_extra(self) -> bool:
        """Return whether setting is a config extra.

        See https://docs.meltano.com/reference/command-line-interface#how-to-use-plugin-extras

        Returns:
            True if setting is a config extra.
        """
        return self.name.startswith("_")

    @property
    def is_custom(self) -> bool:
        """Return whether the setting is custom, i.e. user-defined in `meltano.yml`.

        Returns:
            True if the setting is custom (user defined).
        """
        return self._custom

    @property
    def is_redacted(self) -> bool:
        """Return whether the setting value is redacted.

        Returns:
            True if setting value is redacted.
        """
        return self.kind in {SettingKind.PASSWORD, SettingKind.OAUTH}

    def env_vars(
        self,
        prefixes: list[str],
        include_custom: bool = True,
        for_writing: bool = False,
    ) -> list[EnvVar]:
        """Return environment variables with the provided prefixes.

        Args:
            prefixes: Env var prefixes to prepend.
            include_custom: Include custom env vars from `env_aliases`.
            for_writing: Include target env var from `env`.

        Returns:
            A list of EnvVar instances for this setting definition.
        """
        env_keys = []

        if self.env and for_writing:
            # this ensures we only write to specified `env:`
            env_keys.append(self.env)

        env_keys.extend(utils.to_env_var(prefix, self.name) for prefix in prefixes)
        if not for_writing:
            # read from setting name aliases
            for alias in self.aliases:
                env_keys.extend(utils.to_env_var(prefix, alias) for prefix in prefixes)

        if include_custom:
            env_keys.extend(env for env in self.env_aliases)

        return [EnvVar(key) for key in utils.uniques_in(env_keys)]

    def cast_value(self, value: Any) -> Any:
        """Cast given value.

        Args:
            value: Value to cast.

        Returns:
            Value cast according to specified setting definition kind.

        Raises:
            ValueError: if value cannot be cast to setting kind.
        """
        value = value.isoformat() if isinstance(value, (date, datetime)) else value

        if isinstance(value, str):
            if self.kind == SettingKind.BOOLEAN:
                return utils.truthy(value)
            elif self.kind == SettingKind.INTEGER:
                return int(value)
            elif self.kind == SettingKind.OBJECT:
                value = json.loads(value)
                if not isinstance(value, dict):
                    raise ValueError(f"JSON value '{value}' is not an object")
            elif self.kind == SettingKind.ARRAY:
                value = json.loads(value)
                if not isinstance(value, list):
                    raise ValueError(f"JSON value '{value}' is not an array")

        processor = self.value_processor
        if value is not None and processor:
            if isinstance(processor, str):
                processor = VALUE_PROCESSORS[processor]
            value = processor(value)

        return value

    def post_process_value(self, value: Any) -> Any:
        """Post-process given value.

        Args:
            value: Value to post-process.

        Returns:
            Value post-processed according to any post-processors specified for this
            setting definition.
        """
        processor = self.value_post_processor
        if value is not None and processor:
            if isinstance(processor, str):
                processor = VALUE_PROCESSORS[processor]
            value = processor(value)

        return value

    def stringify_value(self, value: Any) -> str:
        """Return value in string form.

        Args:
            value: Value to stringify

        Returns:
            String form of the passed value.
        """
        if isinstance(value, str):
            return value

        if not self.kind or self.kind == SettingKind.STRING:
            return str(value)

        return json.dumps(value)



"""Meltano run command and supporting functions."""

from __future__ import annotations

import click
import structlog

from meltano.cli import CliError, cli
from meltano.cli.params import pass_project
from meltano.cli.utils import CliEnvironmentBehavior, PartialInstrumentedCmd
from meltano.core.block.blockset import BlockSet
from meltano.core.block.parser import BlockParser, validate_block_sets
from meltano.core.block.plugin_command import PluginCommandBlock
from meltano.core.logging.utils import change_console_log_level
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.runner import RunnerError
from meltano.core.tracking import BlockEvents, Tracker
from meltano.core.tracking.contexts import CliEvent
from meltano.core.tracking.contexts.plugins import PluginsTrackingContext
from meltano.core.utils import click_run_async

logger = structlog.getLogger(__name__)


@cli.command(
    cls=PartialInstrumentedCmd,
    short_help="Run a set of plugins in series.",
    environment_behavior=CliEnvironmentBehavior.environment_required,
)
@click.option(
    "--dry-run",
    help="Do not run, just parse the invocation, validate it, and explain what would be executed.",
    is_flag=True,
)
@click.option(
    "--full-refresh",
    help="Perform a full refresh (ignore state left behind by any previous runs). Applies to all pipelines.",
    is_flag=True,
)
@click.option(
    "--no-state-update",
    help="Run without state saving. Applies to all pipelines.",
    is_flag=True,
)
@click.option(
    "--force",
    "-f",
    help="Force a new run even if a pipeline with the same State ID is already present. Applies to all pipelines.",
    is_flag=True,
)
@click.option(
    "--state-id-suffix",
    help="Define a custom suffix to autogenerate state IDs with.",
)
@click.argument(
    "blocks",
    nargs=-1,
)
@pass_project(migrate=True)
@click.pass_context
@click_run_async
async def run(
    ctx: click.Context,
    project: Project,
    dry_run: bool,
    full_refresh: bool,
    no_state_update: bool,
    force: bool,
    state_id_suffix: str,
    blocks: list[str],
):
    """
    Run a set of command blocks in series.

    Blocks are specified as a list of plugin names, e.g.
    `meltano run some_extractor some_loader some_plugin:some_optional_command` and are run in the order they are specified
    from left to right. A failure in any block will cause the entire run to abort.

    Multiple command blocks can be chained together or repeated, and tap/target pairs will automatically be linked:

        `meltano run tap-gitlab target-postgres dbt:test dbt:run`\n
        `meltano run tap-gitlab target-postgres tap-salesforce target-mysql ...`\n
        `meltano run tap-gitlab target-postgres dbt:run tap-postgres target-bigquery ...`\n

    When running within an active environment, meltano run activates incremental job support. State ID's are autogenerated
    using the format `{active_environment.name}:{extractor_name}-to-{loader_name}(:{state-id-suffix})` for each extract/load pair found:

        `meltano --environment=prod run tap-gitlab target-postgres tap-salesforce target-mysql`\n

    The above command will create two jobs with state IDs `prod:tap-gitlab-to-target-postgres` and `prod:tap-salesforce-to-target-mysql`.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#run
    """
    if dry_run and not ProjectSettingsService.config_override.get("cli.log_level"):
        logger.info("Setting 'console' handler log level to 'debug' for dry run")
        change_console_log_level()

    tracker: Tracker = ctx.obj["tracker"]

    parser_blocks = []  # noqa: F841
    try:
        parser = BlockParser(
            logger,
            project,
            blocks,
            full_refresh,
            no_state_update,
            force,
            state_id_suffix,
        )
        parsed_blocks = list(parser.find_blocks(0))
        if not parsed_blocks:
            tracker.track_command_event(CliEvent.aborted)
            logger.info("No valid blocks found.")
            return
    except Exception as parser_err:
        tracker.track_command_event(CliEvent.aborted)
        raise parser_err

    if validate_block_sets(logger, parsed_blocks):
        logger.debug("All ExtractLoadBlocks validated, starting execution.")
    else:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError("Some ExtractLoadBlocks set failed validation.")
    try:
        await _run_blocks(tracker, parsed_blocks, dry_run=dry_run)
    except Exception as err:
        tracker.track_command_event(CliEvent.failed)
        raise err
    tracker.track_command_event(CliEvent.completed)


async def _run_blocks(
    tracker: Tracker,
    parsed_blocks: list[BlockSet | PluginCommandBlock],
    dry_run: bool,
) -> None:
    for idx, blk in enumerate(parsed_blocks):
        blk_name = blk.__class__.__name__
        tracking_ctx = PluginsTrackingContext.from_block(blk)
        with tracker.with_contexts(tracking_ctx):
            tracker.track_block_event(blk_name, BlockEvents.initialized)
        if dry_run:
            if isinstance(blk, BlockSet):
                logger.info(
                    f"Dry run, but would have run block {idx + 1}/{len(parsed_blocks)}.",
                    block_type=blk_name,
                    comprised_of=[plugin.string_id for plugin in blk.blocks],
                )
            elif isinstance(blk, PluginCommandBlock):
                logger.info(
                    f"Dry run, but would have run block {idx + 1}/{len(parsed_blocks)}.",
                    block_type=blk_name,
                    comprised_of=f"{blk.string_id}:{blk.command}",
                )
            continue

        try:
            await blk.run()
        except RunnerError as err:
            logger.error(
                "Block run completed.",
                set_number=idx,
                block_type=blk_name,
                success=False,
                err=err,
                exit_codes=err.exitcodes,
            )
            with tracker.with_contexts(tracking_ctx):
                tracker.track_block_event(blk_name, BlockEvents.failed)
            raise CliError(
                f"Run invocation could not be completed as block failed: {err}"
            ) from err
        except Exception as bare_err:  # make sure we also fire block failed events for all other exceptions
            with tracker.with_contexts(tracking_ctx):
                tracker.track_block_event(blk_name, BlockEvents.failed)
            raise bare_err

        logger.info(
            "Block run completed.",
            set_number=idx,
            block_type=blk.__class__.__name__,
            success=True,
            err=None,
        )
        with tracker.with_contexts(tracking_ctx):
            tracker.track_block_event(blk_name, BlockEvents.completed)



"""Meltano telemetry contexts for the CLI events."""
from __future__ import annotations

from enum import Enum, auto

import click
from snowplow_tracker import SelfDescribingJson

from meltano.core.tracking.schemas import CliContextSchema
from meltano.core.utils import hash_sha256


class CliEvent(Enum):
    """The kind of event that is occurring in the command-line interface."""

    # The cli command has started, this is fired automatically in most cases when the command is called.
    started = auto()
    # Optionally, a command may fire a `inflight` event to signal that its execution is in progress. This is useful for
    # commands that are long-running and may take a long time to finish, and allows them to emit an event with collected
    # contexts.
    inflight = auto()
    # The cli command has completed without errors.
    completed = auto()
    # Not used in practice.
    skipped = auto()
    # The cli command has failed to complete
    failed = auto()
    # The cli command aborted due to a user induced error.
    aborted = auto()


class CliContext(SelfDescribingJson):
    """CLI context for the Snowplow tracker."""

    def __init__(
        self,
        command: str,
        parent_command_hint: str | None = None,
        options: dict | None = None,
    ):
        """Initialize a CLI context.

        Args:
            command: The command name e.g. `schedule` or `list`.
            parent_command_hint: The upstream parent command e.g. `cli`, `schedule`.
            options: A dict of options keys and sanitized values.
        """
        super().__init__(
            CliContextSchema.url,
            {
                "command": command,
                "parent_command_hint": parent_command_hint,
                "options": options or {},
            },
        )

    @classmethod
    def from_click_context(cls, ctx: click.Context) -> CliContext:
        """Initialize a CLI context.

        Args:
            ctx: The click.Context to derive our tracking context from.

        Returns:
            A CLI context.
        """
        options = {}
        for key, val in ctx.params.items():
            if isinstance(val, (bool, int, float)) or val is None:
                options[key] = val
            else:
                options[key] = hash_sha256(repr(val))

        return cls(
            command=ctx.command.name,
            parent_command_hint=ctx.parent.command.name if ctx.parent else None,
            options=options,
        )



from __future__ import annotations

from meltano.core.plugin import PluginType

from .base import BasePlugin, SingerPlugin
from .catalog import ListExecutor, SelectExecutor
from .mapper import SingerMapper
from .tap import SingerTap
from .target import BookmarkWriter, SingerTarget



"""This module contains the SingerTap class as well as a supporting methods."""

from __future__ import annotations

import asyncio
import json
import logging
import shutil
import sys
from asyncio.streams import StreamReader
from hashlib import sha1
from io import StringIO
from pathlib import Path

import structlog
from jsonschema import Draft4Validator

from meltano.core.behavior.hookable import hook
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.setting_definition import SettingDefinition, SettingKind
from meltano.core.state_service import SINGER_STATE_KEY, StateService
from meltano.core.utils import file_has_data, flatten

from . import PluginType, SingerPlugin
from .catalog import (
    MetadataExecutor,
    MetadataRule,
    SchemaExecutor,
    SchemaRule,
    property_breadcrumb,
    select_filter_metadata_rules,
    select_metadata_rules,
)

logger = structlog.getLogger(__name__)


async def _stream_redirect(
    stream: asyncio.StreamReader, *file_like_objs, write_str=False
):
    """Redirect stream to a file like obj.

    Args:
        stream: the stream to redirect
        file_like_objs: the objects to redirect the stream to
        write_str: if True, stream is written as str
    """
    encoding = sys.getdefaultencoding()
    while not stream.at_eof():
        data = await stream.readline()
        for file_like_obj in file_like_objs:
            file_like_obj.write(data.decode(encoding) if write_str else data)


def _debug_logging_handler(
    name: str, plugin_invoker: PluginInvoker, stderr: StreamReader, *other_dsts
) -> asyncio.Task:
    """Route debug log lines to stderr or an OutputLogger if one is present in our invocation context.

    Args:
        name: name of the plugin
        plugin_invoker: the PluginInvoker to route log lines for
        stderr: stderr StreamReader to route to
        other_dsts: other destinations that the stream should be routed too along with logging output

    Returns:
        asyncio.Task which performs the routing of log lines
    """
    if not plugin_invoker.context or not plugin_invoker.context.base_output_logger:
        return asyncio.ensure_future(
            _stream_redirect(
                stderr, *(sys.stderr, *other_dsts), write_str=True  # noqa: WPS517
            )
        )

    out = plugin_invoker.context.base_output_logger.out(
        name, logger.bind(type="discovery", stdio="stderr")
    )
    with out.line_writer() as outerr:
        return asyncio.ensure_future(
            _stream_redirect(
                stderr, *(outerr, *other_dsts), write_str=True  # noqa: WPS517
            )
        )


def config_metadata_rules(config):  # noqa: WPS210
    """Get metadata rules from config.

    Args:
        config: configuration dict

    Returns:
        a list of MetadataRule
    """
    flat_config = flatten(config, "dot")

    rules = []
    for key, value in flat_config.items():
        # <tap_stream_id>.<key>
        # <tap_stream_id>.<prop>.<key>
        # <tap_stream_id>.<prop>.<subprop>.<key>
        # <tap_stream_id>.properties.<prop>.<key>
        # <tap_stream_id>.properties.<prop>.properties.<subprop>.<key>
        tap_stream_id, *props, rule_key = key.split(".")

        rules.append(
            MetadataRule(
                tap_stream_id=tap_stream_id,
                breadcrumb=property_breadcrumb(props),
                key=rule_key,
                value=value,
            )
        )

    return rules


def config_schema_rules(config):
    """Get schema rules from config.

    Args:
        config: configuration dict

    Returns:
        a list of SchemaRule
    """
    return [
        SchemaRule(
            tap_stream_id=tap_stream_id,
            breadcrumb=["properties", prop],
            payload=payload,
        )
        for tap_stream_id, stream_config in config.items()
        for prop, payload in stream_config.items()
    ]


class SingerTap(SingerPlugin):  # noqa: WPS 214
    """A Plugin for Singer Taps."""

    __plugin_type__ = PluginType.EXTRACTORS

    EXTRA_SETTINGS = [
        SettingDefinition(name="_catalog"),
        SettingDefinition(name="_state"),
        SettingDefinition(name="_load_schema", value="$MELTANO_EXTRACTOR_NAMESPACE"),
        SettingDefinition(name="_select", kind=SettingKind.ARRAY, value=["*.*"]),
        SettingDefinition(
            name="_metadata",
            aliases=["metadata"],
            kind=SettingKind.OBJECT,
            value={},
            value_processor="nest_object",
        ),
        SettingDefinition(
            name="_schema",
            kind=SettingKind.OBJECT,
            value={},
            value_processor="nest_object",
        ),
        SettingDefinition(name="_select_filter", kind=SettingKind.ARRAY, value=[]),
    ]

    def exec_args(self, plugin_invoker):
        """Return the arguments list with the complete runtime paths.

        Args:
            plugin_invoker: the plugin invoker running

        Returns:
            the command line arguments to be passed to the tap
        """
        args = ["--config", plugin_invoker.files["config"]]

        catalog_path = plugin_invoker.files["catalog"]
        if file_has_data(catalog_path):
            if "catalog" in plugin_invoker.capabilities:
                args += ["--catalog", catalog_path]
            elif "properties" in plugin_invoker.capabilities:
                args += ["--properties", catalog_path]
            else:
                logger.warn(
                    "A catalog file was found, but it will be ignored as the extractor does not advertise the `catalog` or `properties` capability"
                )

        state_path = plugin_invoker.files["state"]
        if file_has_data(state_path):
            if "state" in plugin_invoker.capabilities:
                args += ["--state", state_path]
            else:
                logger.warn(
                    "A state file was found, but it will be ignored as the extractor does not advertise the `state` capability"
                )

        return args

    @property
    def config_files(self):
        """Get the configuration files for this tap."""
        return {  # noqa: DAR201
            "config": f"tap.{self.instance_uuid}.config.json",
            "catalog": "tap.properties.json",
            "catalog_cache_key": "tap.properties.cache_key",
            "state": "state.json",
        }

    @property
    def output_files(self):
        """Get the output files for this tap."""
        return {"output": "tap.out"}  # noqa: DAR201

    @hook("before_invoke")
    async def look_up_state_hook(
        self,
        plugin_invoker: PluginInvoker,
        exec_args: tuple[str, ...] = (),
    ):
        """Look up state before being invoked if in sync mode.

        Args:
            plugin_invoker: the plugin invoker running
            exec_args: the args being passed to the tap

        Returns:
            None
        """
        # Use state only in sync mode (i.e. no args)
        if exec_args:
            return

        try:
            await self.look_up_state(plugin_invoker)
        except PluginLacksCapabilityError:
            pass

    async def look_up_state(  # noqa: WPS231, WPS213, WPS210
        self, plugin_invoker: PluginInvoker
    ):
        """Look up state, cleaning up and refreshing as needed.

        Args:
            plugin_invoker: the plugin invoker running

        Returns:
            None

        Raises:
            PluginExecutionError: if state could not be found for this plugin
            PluginLacksCapabilityError: if this plugin does not support incremental state
        """
        if "state" not in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support incremental state"
            )

        state_path = plugin_invoker.files["state"]

        try:
            # Delete state left over from different pipeline run for same extractor
            state_path.unlink()
        except FileNotFoundError:
            pass
        elt_context = plugin_invoker.context
        if not elt_context or not elt_context.job:
            # Running outside pipeline context: incremental state could not be loaded
            return

        if elt_context.full_refresh:
            logger.info(
                "Performing full refresh, ignoring state left behind by any previous runs."
            )
            return

        custom_state_filename = plugin_invoker.plugin_config_extras["_state"]
        if custom_state_filename:
            custom_state_path = plugin_invoker.project.root.joinpath(
                custom_state_filename
            )

            try:
                shutil.copy(custom_state_path, state_path)
                logger.info(f"Found state in {custom_state_filename}")
            except FileNotFoundError as err:
                raise PluginExecutionError(
                    f"Could not find state file {custom_state_path}"
                ) from err

            return
        # the `state.json` is stored in the database
        state = StateService(
            project=elt_context.project, session=elt_context.session
        ).get_state(elt_context.job.job_name)
        if state:
            if state.get(SINGER_STATE_KEY):
                with state_path.open("w") as state_file:
                    json.dump(state.get(SINGER_STATE_KEY), state_file, indent=2)
        else:
            logger.warning("No state was found, complete import.")

    @hook("before_invoke")
    async def discover_catalog_hook(
        self,
        plugin_invoker: PluginInvoker,
        exec_args: tuple[str, ...] = (),
    ):
        """Discover Singer catalog before invoking tap if in sync mode.

        Args:
            plugin_invoker: The invocation handler of the plugin instance.
            exec_args: List of subcommand/args that we where invoked with.

        Returns:
            None
        """
        # Discover only in sync mode (i.e. no args)
        if exec_args:
            return

        try:
            await self.discover_catalog(plugin_invoker)
        except PluginLacksCapabilityError:
            pass

    async def discover_catalog(  # noqa: WPS231, WPS210,
        self, plugin_invoker: PluginInvoker
    ):
        """Perform catalog discovery.

        Args:
            plugin_invoker: The invocation handler of the plugin instance.

        Returns:
            None

        Raises:
            PluginExecutionError: if discovery could not be performed
        """
        catalog_path = plugin_invoker.files["catalog"]
        catalog_cache_key_path = plugin_invoker.files["catalog_cache_key"]
        if catalog_path.exists():
            try:
                cached_key = catalog_cache_key_path.read_text()
                new_cache_key = self.catalog_cache_key(plugin_invoker)

                if cached_key == new_cache_key:
                    logger.debug("Using cached catalog file")
                    return
            except FileNotFoundError:
                pass

            logging.debug("Cached catalog is outdated, running discovery...")

        # We're gonna generate a new catalog, so delete the cache key.
        try:
            catalog_cache_key_path.unlink()
        except FileNotFoundError:
            pass

        custom_catalog_filename = plugin_invoker.plugin_config_extras["_catalog"]
        if custom_catalog_filename:
            custom_catalog_path = plugin_invoker.project.root.joinpath(
                custom_catalog_filename
            )

            try:
                shutil.copy(custom_catalog_path, catalog_path)
                logger.info(f"Found catalog in {custom_catalog_path}")
            except FileNotFoundError as err:
                raise PluginExecutionError(
                    f"Could not find catalog file {custom_catalog_path}"
                ) from err
        else:
            await self.run_discovery(plugin_invoker, catalog_path)

        # test for the result to be a valid catalog
        try:
            with catalog_path.open("r") as catalog_file:
                catalog = json.load(catalog_file)
                Draft4Validator.check_schema(catalog)
        except Exception as err:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Catalog discovery failed: invalid catalog: {err}"
            ) from err

    async def run_discovery(  # noqa: WPS238, WPS210
        self, plugin_invoker: PluginInvoker, catalog_path: Path
    ):  # noqa: DAR401
        """Run tap in discovery mode and store the result.

        Args:
            plugin_invoker: The invocation handler of the plugin instance.
            catalog_path: Where discovery output should be written.

        Raises:
            PluginExecutionError: if state could not be found for this plugin
            PluginLacksCapabilityError: if this plugin does not support incremental state
            Exception: if any other exception occurs
        """
        if "discover" not in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support catalog discovery (the `discover` capability is not advertised)"
            )
        with StringIO("") as stderr_buff:
            try:
                with catalog_path.open(mode="wb") as catalog:
                    handle = await plugin_invoker.invoke_async(
                        "--discover",
                        stdout=asyncio.subprocess.PIPE,
                        stderr=asyncio.subprocess.PIPE,
                        universal_newlines=False,
                    )

                    invoke_futures = [
                        asyncio.ensure_future(_stream_redirect(handle.stdout, catalog)),
                        asyncio.ensure_future(handle.wait()),
                    ]
                    if logger.isEnabledFor(logging.DEBUG) and handle.stderr:
                        invoke_futures.append(
                            _debug_logging_handler(
                                self.name,
                                plugin_invoker,
                                handle.stderr,
                                stderr_buff,
                            )
                        )
                    else:
                        invoke_futures.append(
                            asyncio.ensure_future(
                                _stream_redirect(
                                    handle.stderr, stderr_buff, write_str=True
                                )
                            )
                        )
                    done, _ = await asyncio.wait(
                        invoke_futures,
                        return_when=asyncio.ALL_COMPLETED,
                    )
                    failed = [
                        future for future in done if future.exception() is not None
                    ]
                    if failed:
                        failed_future = failed.pop()
                        raise failed_future.exception()
                exit_code = handle.returncode
            except Exception:
                catalog_path.unlink()
                raise

            if exit_code != 0:
                catalog_path.unlink()
                stderr_buff.seek(0)
                raise PluginExecutionError(
                    f"Catalog discovery failed: command {plugin_invoker.exec_args('--discover')} returned {exit_code} with stderr:\n {stderr_buff.read()}"
                )

    @hook("before_invoke")
    async def apply_catalog_rules_hook(
        self, plugin_invoker: PluginInvoker, exec_args: tuple[str, ...] = ()
    ):
        """Apply catalog rules before invoke if in sync mode.

        Args:
            plugin_invoker: the plugin invoker running
            exec_args: the argumnets to pass to the tap

        Returns:
            None
        """
        # Apply only in sync mode (i.e. no args)
        if exec_args:
            return

        try:
            self.apply_catalog_rules(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def apply_catalog_rules(  # noqa: WPS213, WPS231, WPS210
        self,
        plugin_invoker: PluginInvoker,
        exec_args: tuple[str, ...] = (),
    ):
        """Apply Singer catalog and schema rules to discovered catalog.

        Args:
            plugin_invoker: the plugin invoker running
            exec_args: the argumnets to pass to the tap

        Returns:
            None

        Raises:
            PluginLacksCapabilityError: if plugin does not support entity selection
            PluginExecutionError: if catalog rules could not be applied
        """
        if (
            "catalog" not in plugin_invoker.capabilities
            and "properties" not in plugin_invoker.capabilities
        ):
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support entity selection or catalog metadata and schema rules"
            )

        config = plugin_invoker.plugin_config_extras

        schema_rules = []
        metadata_rules = []

        # If a custom catalog is provided, don't apply catalog rules
        if not config["_catalog"]:
            schema_rules.extend(config_schema_rules(config["_schema"]))

            metadata_rules.extend(select_metadata_rules(["!*.*"]))
            metadata_rules.extend(select_metadata_rules(config["_select"]))
            metadata_rules.extend(config_metadata_rules(config["_metadata"]))

        # Always apply select filters (`meltano elt` `--select` and `--exclude` options)
        metadata_rules.extend(select_filter_metadata_rules(config["_select_filter"]))

        if not schema_rules and not metadata_rules:
            return

        catalog_path = plugin_invoker.files["catalog"]
        catalog_cache_key_path = plugin_invoker.files["catalog_cache_key"]

        try:
            with catalog_path.open() as catalog_file:
                catalog = json.load(catalog_file)

            if schema_rules:
                SchemaExecutor(schema_rules).visit(catalog)

            if metadata_rules:
                MetadataExecutor(metadata_rules).visit(catalog)

            with catalog_path.open("w") as catalog_f:
                json.dump(catalog, catalog_f, indent=2)

            cache_key = self.catalog_cache_key(plugin_invoker)
            if cache_key:
                catalog_cache_key_path.write_text(cache_key)
            else:
                try:  # noqa: WPS505
                    catalog_cache_key_path.unlink()
                except FileNotFoundError:
                    pass
        except FileNotFoundError as err:
            raise PluginExecutionError(
                "Applying catalog rules failed: catalog file is missing."
            ) from err
        except Exception as err:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Applying catalog rules failed: catalog file is invalid: {err}"
            ) from err

    def catalog_cache_key(self, plugin_invoker):
        """Get a cache key for the catalog.

        Args:
            plugin_invoker: the plugin invoker running

        Returns:
            the cache key for the catalog, if plugin catalog can be cached
        """
        # Treat non-pip plugins as editable/dev-mode plugins and do not cache.
        if plugin_invoker.plugin.pip_url is None:
            return None

        # If the extractor is installed as editable, don't cache because
        # the result of discovery could change at any time.
        if plugin_invoker.plugin.pip_url.startswith("-e"):
            return None

        extras = plugin_invoker.plugin_config_extras

        # If a custom catalog is provided, there's no need to cache
        if extras["_catalog"]:
            return None

        # The catalog should be regenerated, and the catalog cache invalidated,
        # if any settings changed that could affect discovery, or if schema or
        # metadata rules changed.
        # Changes to selection rules and selection filter rules are ignored,
        # since "selected" metadata is reset using the `!*.*` selection rule anyway.
        key_dict = {
            **plugin_invoker.plugin_config,
            "_schema": extras["_schema"],
            "_metadata": extras["_metadata"],
        }

        key_json = json.dumps(key_dict)

        return sha1(key_json.encode()).hexdigest()  # noqa: S303 S324
